#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, List, Callable
import os
import json
import pickle
import pandas as pd
import numpy as np
from mgktools.evaluators.metric import Metric
from molalkit.active_learning.selector import BaseSelector, RandomSelector
from molalkit.active_learning.forgetter import BaseForgetter, RandomForgetter, FirstForgetter
from molalkit.active_learning.utils import eval_metric_func
from molalkit.data.utils import get_subset_from_uidx


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


class ActiveLearningResult:
    def __init__(self, n_iter: int):
        self.n_iter = n_iter
        self.uidx_before = []
        self.uidx_select = []
        self.acquisition_select = []
        self.uidx_forget = []
        self.acquisition_forget = []
        self.uidx_after = []
        self.performance = dict()


class ActiveLearningTrajectory:
    def __init__(self, metrics: List[str] = None):
        self.metrics = metrics
        self.results = []

    def update_performance(self, column_name: str, results_dict: Dict) -> bool:
        results = [alr.performance.get(column_name) for alr in self.results]
        if results[0] is not None:
            results_dict[column_name] = results
            return True
        else:
            return False

    def get_results(self) -> Dict:
        results_dict = dict()
        results_dict["n_iter"] = [alr.n_iter for alr in self.results]
        # collect the results of various metrics during active learning
        if self.metrics is not None:
            for metric in self.metrics:
                for i in range(100):
                    tag = self.update_performance(f"{metric}-model_{i}", results_dict)
                    if not tag:
                        break
        # collect the results of top data selected during active learning
        self.update_performance("top_score", results_dict)
        results_dict["uidx_before"] = [json.dumps(alr.uidx_before, cls=NpEncoder) for alr in self.results]
        results_dict["uidx_select"] = [json.dumps(alr.uidx_select, cls=NpEncoder) for alr in self.results]
        results_dict["acquisition_select"] = [json.dumps(alr.acquisition_select, cls=NpEncoder) for alr in self.results]
        results_dict["uidx_forget"] = [json.dumps(alr.uidx_forget, cls=NpEncoder) for alr in self.results]
        results_dict["acquisition_forget"] = [json.dumps(alr.acquisition_forget, cls=NpEncoder) for alr in self.results]
        results_dict["uidx_after"] = [json.dumps(alr.uidx_after, cls=NpEncoder) for alr in self.results]
        return results_dict


class ActiveLearner:
    def __init__(self, save_dir: str, selector: BaseSelector, forgetter: BaseForgetter,
                 models, id2datapoints: List[Dict],
                 datasets_train, datasets_pool, datasets_val = None,
                 metrics: List[Metric] = None, top_uidx: List[int] = None,
                 kernel: Callable = None, detail: bool = False):
        """ Active learning class.
        
        Parameters
        save_dir: str
            Directory to save the results of active learning.
        selector: BaseSelector
            Selection method used to select the data from the pool set.
        forgetter: BaseForgetter
            Forget method used to forget the data from the training set.
        models:
            ML models used in active learning. [selector, evaluator1, evaluator2, ...]
        datasets_train:
            Training datasets used to train the model. [selector, evaluator1, evaluator2, ...]
        dataset_pool:
            Pool set used to select the data during active learning.
        dataset_val:
            Validation set used to evaluate the model performance.
        metrics: List[Metric] = None
            Metrics used to evaluate the ML model performance. This is usually used in explorative 
            active learning in which the goal is to build a model with the best performance.
        top_uidx: List[int] = None
            List of data index that are considered as top data. If this is given, the percentage of 
            top data selected by active learning will be outputed. This is a metric that usually used 
            in exploitive active learning.
        kernel: Callable = None
            Kernel function used in cluster selection method.
        detail: bool = False
            If True, the detailed information of the model performance at each iteration will be outputed.
        """
        self.save_dir = save_dir
        self.selector = selector
        self.forgetter = forgetter
        self.models = models
        self.id2datapoints = id2datapoints
        self.datasets_train = datasets_train
        self.datasets_pool = datasets_pool
        self.datasets_val = datasets_val
        # input for model performance evaluation
        self.metrics = metrics
        self.top_uidx = top_uidx
        # kernel used for cluster selection method
        self.kernel = kernel
        self.detail = detail
        # initialize the active learning trajectory
        self.current_iter = 0
        self.model_fitted = False
        self.active_learning_traj = ActiveLearningTrajectory(metrics=self.metrics)

    @property
    def train_size(self) -> int:
        return len(self.datasets_train[0])

    @property
    def val_size(self) -> int:
        return 0 if self.datasets_val is None else len(self.datasets_val[0])

    def step_select(self, stop_cutoff: float = None, confidence_cutoff: float = None):
        alr = ActiveLearningResult(self.current_iter)
        alr.uidx_before = [data.uidx for data in self.datasets_train[0]]
        assert self.selector is not None, "You need to provide a selector before step_select()."
        # train the model if it is not trained in the evaluation step, and the selection method is not random.
        if not self.model_fitted and not isinstance(self.selector, RandomSelector):
            self.models[0].fit_molalkit(self.datasets_train[0], iteration=self.current_iter)
        selected_idx, acquisition, remain_idx = self.selector(model=self.models[0],
                                                              dataset_pool=self.datasets_pool[0],
                                                              kernel=self.kernel,
                                                              stop_cutoff=stop_cutoff,
                                                              confidence_cutoff=confidence_cutoff)
        alr.uidx_select = [self.datasets_pool[0].data[idx].uidx for idx in selected_idx]
        alr.acquisition_select = acquisition
        # transfer data from pool to train.
        alr.uidx_after = alr.uidx_before + alr.uidx_select
        uidx_pool = [self.datasets_pool[0].data[idx].uidx for idx in remain_idx]
        ### update the dataset_train and dataset_pool
        for i, id2datapoint in enumerate(self.id2datapoints):
            self.datasets_train[i] = get_subset_from_uidx(self.datasets_train[i], id2datapoint, alr.uidx_after)
            self.datasets_pool[i] = get_subset_from_uidx(self.datasets_pool[i], id2datapoint, uidx_pool)
        # set the model unfitted because new data is added.
        self.model_fitted = False
        self.current_iter += 1
        self.active_learning_traj.results.append(alr)

    def step_forget(self, stop_cutoff: float = None, confidence_cutoff: float = None):
        alr = ActiveLearningResult(self.current_iter)
        alr.uidx_before = [data.uidx for data in self.datasets_train[0]]
        assert self.forgetter is not None, "You need to provide a forgetter before step_forget()."
        # train the model if the forgetter is not random or first.
        if not self.model_fitted and not self.forgetter.__class__ in [RandomForgetter, FirstForgetter]:
            self.models[0].fit_molalkit(self.datasets_train[0], iteration=self.current_iter)
        # forget algorithm is applied.
        forget_idx, acquisition, remain_idx = self.forgetter(model=self.models[0],
                                                             dataset_train=self.datasets_train[0],
                                                             stop_cutoff=stop_cutoff,
                                                             confidence_cutoff=confidence_cutoff)
        alr.uidx_forget = [self.datasets_train[0].data[i].uidx for i in forget_idx]
        alr.acquisition_forget = acquisition
        # transfer data from train to pool.
        alr.uidx_after = [self.datasets_train[0].data[i].uidx for i in remain_idx]# alr.uidx_before + alr.uidx_select
        uidx_pool = [data.uidx for data in self.datasets_pool[0]] + alr.uidx_forget
        ### update the dataset_train and dataset_pool
        for i, id2datapoint in enumerate(self.id2datapoints):
            self.datasets_train[i] = get_subset_from_uidx(self.datasets_train[i], id2datapoint, alr.uidx_after)
            self.datasets_pool[i] = get_subset_from_uidx(self.datasets_pool[i], id2datapoint, uidx_pool)
        # set the model unfitted because new data is added.
        self.model_fitted = False
        self.current_iter += 1
        self.active_learning_traj.results.append(alr)

    def evaluate(self):
        if len(self.active_learning_traj.results) == 0:
            alr = ActiveLearningResult(self.current_iter - 1)
            alr.uidx_before = []
            alr.uidx_after = [data.uidx for data in self.datasets_train[0]]
            self.active_learning_traj.results.append(alr)
        else:
            alr = self.active_learning_traj.results[-1]
        # evaluate the prediction performance of ML model on the validation set
        if self.metrics is not None:
            for i, model in enumerate(self.models):
                model.fit_molalkit(self.datasets_train[i])
                y_pred = np.asarray(model.predict_value(self.datasets_val[i])).ravel()
                y_pred = [data.targets[0] if data.uidx in alr.uidx_after else y_pred[i] for i, data in enumerate(self.datasets_val[i])]
                if self.detail:
                    df = pd.DataFrame({"true": self.datasets_val[i].y.ravel(), "pred": y_pred})
                    df.to_csv(os.path.join(self.save_dir, f"model_{i}-iter_{self.current_iter}.csv"), index=False)
                for metric in self.metrics:
                    metric_value = eval_metric_func(self.datasets_val[i].y.ravel(), y_pred, metric=metric)
                    alr.performance[f"{metric}-model_{i}"] = metric_value
            self.model_fitted = True
        # evaluate the percentage of top data selected in the training set
        if self.top_uidx is not None:
            alr.performance["top_score"] = self.get_top_score(self.datasets_train[0], self.top_uidx)

    def write_traj(self):
        df_traj = pd.DataFrame(self.active_learning_traj.get_results())
        df_traj.to_csv(os.path.join(self.save_dir, "al_traj.csv"), index=False)

    @staticmethod
    def get_top_score(dataset, top_uidx) -> float:
        N_top_k = 0
        for data in dataset:
            if data.uidx in top_uidx:
                N_top_k += 1
        return N_top_k / len(top_uidx)

    def save(self, path, filename="al.pkl", overwrite=False):
        f_al = os.path.join(path, filename)
        if os.path.isfile(f_al) and not overwrite:
            raise RuntimeError(
                f"Path {f_al} already exists. To overwrite, set "
                "`overwrite=True`."
            )
        store = self.__dict__.copy()
        # Chemprop TrainArgs is unpicklable, transform into dict.
        for model in store["models"]:
            if hasattr(model, "chemprop_train_args"):
                model.chemprop_train_args = model.chemprop_train_args.as_dict()
                model.chemprop_predict_args = model.chemprop_predict_args.as_dict()
        pickle.dump(store, open(f_al, "wb"), protocol=4)
        # transform back to TrainArgs
        for model in store["models"]:
            if hasattr(model, "chemprop_train_args"):
                from chemprop.args import TrainArgs, PredictArgs
                model.chemprop_train_args = TrainArgs().from_dict(model.chemprop_train_args, skip_unsettable=True)
                model.chemprop_predict_args = PredictArgs().from_dict(model.chemprop_predict_args, skip_unsettable=True)
                # model.save_checkpoint(), Not needed as the model weights are already saved in the active learner pickle file.

    @classmethod
    def load(cls, path, filename="al.pkl"):
        f_al = os.path.join(path, filename)
        store = pickle.load(open(f_al, "rb"))
        # transform Chemprop TrainArgs from dict back to TrainArgs
        for model in store["models"]:
            if hasattr(model, "chemprop_train_args"):
                from chemprop.args import TrainArgs, PredictArgs
                model.chemprop_train_args = TrainArgs().from_dict(model.chemprop_train_args, skip_unsettable=True)
                model.chemprop_predict_args = PredictArgs().from_dict(model.chemprop_predict_args, skip_unsettable=True)
                # model.load_checkpoint(), Not needed as the model weights are already saved in the active learner pickle file.
        input = {}
        for key in ["save_dir", "selector", "forgetter", "models", 
                    "id2datapoints", "datasets_train", "datasets_pool"]:
            input[key] = store[key]
        learner = cls(**input)
        learner.__dict__.update(**store)
        return learner



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from sklearn.svm import SVR
from molalkit.models.base import BaseSklearnModel


class SVRegressor(SVR, BaseSklearnModel):
    def fit_molalkit(self, train_data, iteration: int = 0):
        return self.fit_molalkit_(train_data, self)

    def predict_uncertainty(self, pred_data):
        raise ValueError("SVRegressor does not support uncertainty prediction")

    def predict_value(self, pred_data):
        X = pred_data.X
        return super().predict(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
import numpy as np


class BaseModel(ABC):
    @abstractmethod
    def fit_molalkit(self, data, iteration: int = 0):
        pass

    @abstractmethod
    def predict_value(self, data):
        pass

    @abstractmethod
    def predict_uncertainty(self, data):
        pass


class BaseSklearnModel(BaseModel, ABC):
    @staticmethod
    def fit_molalkit_(train_data, sklearn_model):
        X = train_data.X
        y = train_data.y
        assert y.ndim == 2
        assert y.shape[1] == 1
        y = y.ravel()
        return sklearn_model.fit(X, y)

    @staticmethod
    def predict_uncertainty_c(pred_data, model):
        X = pred_data.X
        p = model.predict_proba(X)
        # cross entropy
        # return entropy(p, base=2, axis=1)
        return (0.25 - np.var(p, axis=1)) * 4

    @staticmethod
    def predict_value_c(pred_data, model):
        X = pred_data.X
        return model.predict_proba(X)[:, 1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import List, Tuple, Union
from torch.utils.data import Dataset
from molalkit.models.rnn.tokenizer import BaseTokenizer


class MoleculeDataset(Dataset):
    """Dataset class for molecular data."""
    
    def __init__(self, smiles_list: List[str],
                 targets: List[Union[int, float]],
                 tokenizer: BaseTokenizer):
        self.smiles_list = smiles_list
        self.targets = targets
        self.tokenizer = tokenizer
        self.data = [tokenizer.encode(smiles, 
                                      generative=False, 
                                      padding=False) for smiles in smiles_list]

    def __len__(self) -> int:
        return len(self.smiles_list)
    
    def __getitem__(self, index: int) -> Tuple[List[int], Union[int, float]]:
        return self.data[index], self.targets[index]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import List, Tuple
import numpy as np
from molalkit.active_learning.utils import random_choice, get_topn_idx
from molalkit.models.random_forest.RandomForestClassifier import RFClassifier
from molalkit.models.gaussian_process.GaussianProcessRegressor import GPRegressor


class BaseForgetter(ABC):
    def __init__(self, batch_size: int = 1):
        self.batch_size = batch_size

    @abstractmethod
    def __call__(self, **kwargs) -> Tuple[List[int], List[float]]:
        pass

    @property
    @abstractmethod
    def info(self) -> str:
        pass


class BaseRandomForgetter(BaseForgetter, ABC):
    """Base Forgetter that uses random seed."""
    def __init__(self, batch_size: int = 1, seed: int = 0):
        super().__init__(batch_size=batch_size)
        np.random.seed(seed)


class RandomForgetter(BaseRandomForgetter):
    def __call__(self, dataset_train, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        idx, idx_remain = random_choice(len(dataset_train), self.batch_size)
        return idx, [], idx_remain

    @property
    def info(self) -> str:
        return f"RandomForgetter(batch_size={self.batch_size})"


class FirstForgetter(BaseForgetter):
    def __call__(self, dataset_train, **kwargs) -> Tuple[List[int], None]:
        return list(range(self.batch_size)), [], list(range(self.batch_size, len(dataset_train)))

    @property
    def info(self) -> str:
        return f"FirstForgetter(batch_size={self.batch_size})"


class MinOOBUncertaintyForgetter(BaseRandomForgetter):
    """ Forget the samples with the lowest out-of-bag (OOB) uncertainty. 
        This method is only applicable to random forest classifier. 
    """
    def __call__(self, model: RFClassifier, dataset_train, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, RFClassifier), "Only random forest classifier is supported for OOB prediction."
        assert model.oob_score is True, "You should set oob_score=True when training the model."
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        y_oob_uncertainty = (0.25 - np.var(y_oob_proba, axis=1)) * 4
        # select the top-n points with least uncertainty
        idx = get_topn_idx(y_oob_uncertainty, n=self.batch_size, target="min")

        acquisition = y_oob_uncertainty[np.array(idx)].tolist()

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MinOOBUncertaintyForgetter(batch_size={self.batch_size})"


class MaxOOBUncertaintyForgetter(BaseRandomForgetter):
    """ Forget the samples with the highest out-of-bag (OOB) uncertainty.
        This method is only applicable to random forest classifier.
    """
    def __call__(self, model: RFClassifier, dataset_train, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, RFClassifier), "Only random forest classifier is supported for OOB prediction."
        assert model.oob_score is True, "You should set oob_score=True when training the model."
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        y_oob_uncertainty = (0.25 - np.var(y_oob_proba, axis=1)) * 4
        # select the top-n points with least uncertainty
        idx = get_topn_idx(y_oob_uncertainty, n=self.batch_size, target="max")

        acquisition = y_oob_uncertainty[np.array(idx)].tolist()

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MaxOOBUncertaintyForgetter(batch_size={self.batch_size})"


class MinOOBErrorForgetter(BaseRandomForgetter):
    """ Forget the samples with the lowest out-of-bag (OOB) error.
        This method is only applicable to random forest classifier.
    """
    def __call__(self, model: RFClassifier, dataset_train, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, RFClassifier), "Only random forest classifier is supported for OOB prediction."
        assert model.oob_score is True, "You should set oob_score=True when training the model."
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        oob_error = np.absolute(y_oob_proba[:, 1] - dataset_train.y.ravel())
        # select the top-n points with least uncertainty
        idx = get_topn_idx(oob_error, n=self.batch_size, target="min")

        acquisition = oob_error[np.array(idx)].tolist() if idx else []

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MinOOBErrorForgetter(batch_size={self.batch_size})"


class MaxOOBErrorForgetter(BaseRandomForgetter):
    """ Forget the samples with the highest out-of-bag (OOB) error.
        This method is only applicable to random forest classifier.
    """
    def __call__(self, model: RFClassifier, dataset_train, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, RFClassifier), "Only random forest classifier is supported for OOB prediction."
        assert model.oob_score is True, "You should set oob_score=True when training the model."
        y_oob_proba = model.oob_decision_function_
        # uncertainty calculation, normalized into 0 to 1
        oob_error = np.absolute(y_oob_proba[:, 1] - dataset_train.y.ravel())
        # select the top-n points with least uncertainty
        idx = get_topn_idx(oob_error, n=self.batch_size, target="max")

        acquisition = oob_error[np.array(idx)].tolist() if idx else []

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MaxOOBErrorForgetter(batch_size={self.batch_size})"


class MinLOOUncertaintyForgetter(BaseRandomForgetter):
    def __call__(self, model: GPRegressor, dataset_train, **kwargs) -> Tuple[List[int], List[float]]:
        """ Forget the samples with the lowest Leave-one-out cross-validation (LOOCV) uncertainty.
        Parameters
        ----------
        model: Only Gaussian process regressor is supported due to efficient LOOCV of GPR.
        dataset_train: The dataset to forget.

        Returns
        -------
        The index and the acquisition value of samples to forget.
        """
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, GPRegressor), "Only Gaussian process regressor is supported for LOOCV prediction."
        y_loocv, y_loocv_std = model.predict_loocv(dataset_train.X, dataset_train.y.ravel(), return_std=True)
        # uncertainty calculation, normalized into 0 to 1
        loo_uncertainty = y_loocv_std
        # select the top-n points with least uncertainty
        idx = get_topn_idx(loo_uncertainty, n=self.batch_size, target="min")

        acquisition = loo_uncertainty[np.array(idx)].tolist() if idx else []

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MinLOOUncertaintyForgetter(batch_size={self.batch_size})"


class MaxLOOUncertaintyForgetter(BaseRandomForgetter):
    def __call__(self, model: GPRegressor, dataset_train, **kwargs) -> Tuple[List[int], List[float]]:
        """ Forget the samples with the highest Leave-one-out cross-validation (LOOCV) uncertainty.
        Parameters
        ----------
        model: Only Gaussian process regressor is supported due to efficient LOOCV of GPR.
        dataset_train: The dataset to forget.

        Returns
        -------
        The index and the acquisition value of samples to forget.
        """
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, GPRegressor), "Only Gaussian process regressor is supported for LOOCV prediction."
        y_loocv, y_loocv_std = model.predict_loocv(dataset_train.X, dataset_train.y.ravel(), return_std=True)
        # uncertainty calculation, normalized into 0 to 1
        loo_uncertainty = y_loocv_std
        # select the top-n points with least uncertainty
        idx = get_topn_idx(loo_uncertainty, n=self.batch_size, target="max")

        acquisition = loo_uncertainty[np.array(idx)].tolist() if idx else []

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MaxLOOUncertaintyForgetter(batch_size={self.batch_size})"


class MinLOOErrorForgetter(BaseRandomForgetter):
    def __call__(self, model: GPRegressor, dataset_train, **kwargs) -> Tuple[List[int], List[float]]:
        """ Forget the samples with the lowest Leave-one-out cross-validation (LOOCV) error.
        Parameters
        ----------
        model: Only Gaussian process regressor is supported due to efficient LOOCV of GPR.
        dataset_train: The dataset to forget.

        Returns
        -------
        The index and the acquisition value of samples to forget.
        """
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, GPRegressor), "Only Gaussian process regressor is supported for LOOCV prediction."
        y_loocv = model.predict_loocv(dataset_train.X, dataset_train.y.ravel(), return_std=False)
        # uncertainty calculation, normalized into 0 to 1
        loo_error = np.absolute(y_loocv - dataset_train.y.ravel())
        # select the top-n points with least uncertainty
        idx = get_topn_idx(loo_error, n=self.batch_size, target="min")

        acquisition = loo_error[np.array(idx)].tolist() if idx else []

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MinOOBErrorForgetter(batch_size={self.batch_size})"


class MaxLOOErrorForgetter(BaseRandomForgetter):
    def __call__(self, model: GPRegressor, dataset_train, **kwargs) -> Tuple[List[int], List[float]]:
        """ Forget the samples with the highest Leave-one-out cross-validation (LOOCV) error.
        Parameters
        ----------
        model: Only Gaussian process regressor is supported due to efficient LOOCV of GPR.
        dataset_train: The dataset to forget.

        Returns
        -------
        The index and the acquisition value of samples to forget.
        """
        # assert self.batch_size < len(dataset_train), "batch_size should be less than the size of the dataset."
        assert isinstance(model, GPRegressor), "Only Gaussian process regressor is supported for LOOCV prediction."
        y_loocv = model.predict_loocv(dataset_train.X, dataset_train.y.ravel(), return_std=False)
        # uncertainty calculation, normalized into 0 to 1
        loo_error = np.absolute(y_loocv - dataset_train.y.ravel())
        # select the top-n points with least uncertainty
        idx = get_topn_idx(loo_error, n=self.batch_size, target="max")

        acquisition = loo_error[np.array(idx)].tolist() if idx else []

        mask = np.isin(range(len(dataset_train)), idx)
        idx_remain = np.arange(len(dataset_train))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"MaxLOOErrorForgetter(batch_size={self.batch_size})"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import List, Dict, Set, Optional
import re
import selfies as sf


class BaseTokenizer(ABC):
    """
    Abstract base class for tokenizers.

    Methods
    -------
    tokenize(smiles: str) -> List[str]
        Tokenizes a SMILES string.
    create_vocabulary(smiles_list: List[str]) -> Dict[str, int]
        Creates a vocabulary from a list of SMILES strings.
    encode(smiles: str) -> List[int]
        Encodes a SMILES string into a list of integers.
    """

    @abstractmethod
    def tokenize(self, smiles: str) -> List[str]:
        pass

    @abstractmethod
    def create_vocabulary(self, smiles_list: List[str]) -> Dict[str, int]:
        pass
    
    @abstractmethod
    def encode(self, smiles: str) -> List[int]:
        pass


class SMILESTokenizer(BaseTokenizer):
    """
    A simple regex-based tokenizer adapted from the deepchem smiles_tokenizer package.
    SMILES regex pattern for the tokenization is designed by Schwaller et. al., ACS Cent. Sci 5 (2019).

    Attributes
    ----------
    regex_pattern : str
        The regex pattern used for tokenizing SMILES strings.
    regex : re.Pattern
        The compiled regex pattern.
    special_tokens : List[str]
        List of special tokens used in tokenization.

    Methods
    -------
    tokenize(smiles: str) -> List[str]
        Tokenizes a SMILES string.
    create_vocabulary(smiles_list: List[str]) -> Dict[str, int]
        Creates a vocabulary from a list of SMILES strings.
    encode(smiles: str) -> List[int]
        Encodes a SMILES string into a list of integers.
    """

    def __init__(self):
        """
        Initializes the SMILESTokenizer with a regex pattern and special tokens.
        """
        self.regex_pattern = (
            r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\."
            r"|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
        )
        self.regex = re.compile(self.regex_pattern)
        self.special_tokens = ['<eos>', '<sos>', '<pad>']

    def tokenize(self, smiles):
        """
        Tokenizes SMILES string.

        Parameters
        ----------
        smiles : str
            Input SMILES string.

        Returns
        -------
        List[str]
            A list of tokens.
        """
        tokens = [token for token in self.regex.findall(smiles)]
        return tokens

    def create_vocabulary(self, smiles_list: List[str]) -> Dict[str, int]:
        """
        Create a vocabulary dictionary from a list of SMILES strings.

        Parameters:
        ----------
            smiles_list (List[str]): List of SMILES strings to process.

        Returns
        -------
        Dict[str, int]
            Dictionary mapping tokens to indices.
        """
        assert not hasattr(self, "vocab"), "Vocabulary already exists. Create a new tokenizer instead."
        # Collect all unique tokens
        tokens = set()
        max_length = 0
        for smiles in smiles_list:
            tokens_ = self.tokenize(smiles)
            tokens.update(tokens_)
            if len(tokens) > max_length:
                max_length = len(tokens)
        
        # Create vocabulary with indices
        tokens = list(sorted(tokens)) + list(self.special_tokens)
        vocab_dict = {token: idx for idx, token in enumerate(tokens)}

        self.vocab = vocab_dict
        self.max_length = max_length
        return vocab_dict

    def encode(self, smiles: str, generative: bool = False, padding: bool = True) -> List[int]:
        """
        Encode a SMILES string as a list of token indices.

        Parameters
        ----------
            smiles (str): Input SMILES string.

        Returns
        -------
        List[int]
            List of token indices.
        """
        assert hasattr(self, "vocab"), "Vocabulary not created. Call create_vocabulary() first."
        tokens = self.tokenize(smiles)
        ints = [self.vocab[token] for token in tokens]
        if generative:
            ints = [self.vocab['<sos>'], *ints, self.vocab['<eos>']]
            max_length = self.max_length + 2
        else:
            max_length = self.max_length
        if padding:
            ints += [self.vocab['<pad>']] * (max_length - len(ints))

        return ints


class SELFIESTokenizer(BaseTokenizer):
    """
    A tokenizer for converting SMILES to SELFIES and generating vocabulary.
    Uses the SELFIES library for molecular string representation.
    """

    def __init__(self):
        """
        Initializes the SELFIESTokenizer with special tokens.
        """
        self.special_tokens = ['.', '<eos>', '<sos>', '<pad>']

    def tokenize(self, smiles: str) -> List[str]:
        """
        Tokenizes SMILES string.

        Parameters
        ----------
        smiles : str
            Input SMILES string.

        Returns
        -------
        List[str]
            A list of tokens.
        """
        selfies = sf.encoder(smiles)
        return list(sf.split_selfies(selfies))

    def create_vocabulary(self, smiles_list: List[str]) -> Dict[str, int]:
        """
        Create a vocabulary dictionary from a list of SMILES strings.

        Parameters:
        ----------
            smiles_list (List[str]): List of SMILES strings to process.

        Returns
        -------
        Dict[str, int]
            Dictionary mapping tokens to indices.
        """
        assert not hasattr(self, "vocab"), "Vocabulary already exists. Create a new tokenizer instead."
        # Collect all unique tokens
        tokens = set()
        max_length = 0
        for smiles in smiles_list:
            tokens_ = self.tokenize(smiles)
            tokens.update(tokens_)
            if len(tokens) > max_length:
                max_length = len(tokens)
        
        # Create vocabulary with indices
        tokens = list(sorted(tokens)) + list(self.special_tokens)
        vocab_dict = {token: idx for idx, token in enumerate(tokens)}

        self.vocab = vocab_dict
        self.max_length = max_length
        return vocab_dict

    def encode(self, smiles: str, generative: bool = False, padding: bool = True) -> List[int]:
        """
        Encode a SMILES string as a list of token indices.

        Parameters
        ----------
            smiles (str): Input SMILES string.

        Returns
        -------
        List[int]
            List of token indices.
        """

        assert hasattr(self, "vocab"), "Vocabulary not created. Call create_vocabulary() first."
        tokens = self.tokenize(smiles)
        ints = [self.vocab[token] for token in tokens]
        if generative:
            ints = [self.vocab['<sos>'], *ints, self.vocab['<eos>']]
            max_length = self.max_length + 2
        else:
            max_length = self.max_length
        if padding:
            ints += [self.vocab['<pad>']] * (max_length - len(ints))

        return ints



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import List, Tuple, Callable
import numpy as np
from collections import defaultdict
from sklearn.cluster import KMeans
from sklearn.manifold import SpectralEmbedding
from molalkit.active_learning.utils import random_choice, get_topn_idx
from molalkit.data.utils import get_subset_from_idx


class BaseSelector(ABC):
    def __init__(self, batch_size: int = 1):
        self.batch_size = batch_size

    @abstractmethod
    def __call__(self, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        pass

    @property
    @abstractmethod
    def info(self) -> str:
        pass


class BaseRandomSelector(BaseSelector, ABC):
    """Base Selector that uses random seed."""
    def __init__(self, batch_size: int = 1, seed: int = 0):
        super().__init__(batch_size=batch_size)
        self.seed = seed
        np.random.seed(seed)


class BaseClusterSelector(BaseRandomSelector, ABC):
    """ Base Selector that uses clustering method.
        cluster_size samples will be selected first according to provided selection algorithm,
        and then separate into batch_size clusters using KMeans clustering method.
        Finally, one sample from each cluster will be selected.
        This can increase the diversity of the selected samples batch.
    """
    def __init__(self, batch_size: int = 1, seed: int = 0, cluster_size: int = None):
        super().__init__(batch_size=batch_size, seed=seed)
        assert batch_size > 1, "batch_size should be larger than 1 for cluster selection method."
        if cluster_size is None:
            self.cluster_size = batch_size * 20
        else:
            assert cluster_size >= batch_size, "cluster_size should be larger than batch_size."
            self.cluster_size = cluster_size

    def get_idx_cluster(self, dataset_pool, kernel: Callable, idx_cluster) -> List[int]:
        """ Find distant samples from a pool using KMeans clustering method."""
        if len(idx_cluster) < self.batch_size:
            return list(range(len(idx_cluster)))
        else:
            K = kernel(dataset_pool.X[idx_cluster])
            add_idx = self.find_distant_samples(gram_matrix=K, batch_size=self.batch_size)
            return add_idx

    def find_distant_samples(self, gram_matrix: np.ndarray[float], batch_size: int = 1) -> List[int]:
        """ Find distant samples from a pool using clustering method.

        Parameters
        ----------
        gram_matrix: gram (kernel) matrix of the samples.
        batch_size: number of samples to be selected.

        Returns
        -------
        List of idx
        """
        embedding = SpectralEmbedding(
            n_components=batch_size,
            affinity="precomputed",
            random_state=self.seed
        ).fit_transform(gram_matrix)

        cluster_result = KMeans(
            n_clusters=batch_size,
            random_state=self.seed
        ).fit_predict(embedding)
       # Calculate cluster centers
        centers = np.array([embedding[cluster_result == i].mean(axis=0)
                           for i in range(batch_size)])
        # For each cluster, find the sample that's most distant from other cluster centers
        total_distance = defaultdict(dict)
        for i in range(batch_size):
            cluster_idx = cluster_result[i]
            # Get centers of other clusters
            other_centers = np.delete(centers, cluster_idx, axis=0)
            # Calculate sum of inverse distances to other centers
            # (Using inverse distance means smaller values indicate more distant points)
            sum_inverse_distances = np.sum(np.sqrt(np.sum(
                np.square(embedding[i] - other_centers), axis=1
            )) ** -0.5)
            total_distance[cluster_idx][sum_inverse_distances] = i
        # Select the most distant sample from each cluster
        selected_indices = [
            total_distance[i][min(total_distance[i].keys())]
            for i in range(batch_size)
        ]
        return selected_indices


class BasePartialQuerySelector(BaseRandomSelector, ABC):
    """ Base Selector that uses partial query method.
        query_size samples will be selected first randomly,
        and then batch_size samples will be selected according to provided selection algorithm.
    """
    @staticmethod
    def get_partial_data(dataset_pool, query_size):
        # randomly select query_size samples from dataset_pool.
        if len(dataset_pool) <= query_size:
            return dataset_pool, np.array(range(len(dataset_pool)))
        else:
            idx_query = np.random.choice(range(len(dataset_pool)), query_size, replace=False)
            data_query = get_subset_from_idx(dataset_pool, idx_query)
            return data_query, idx_query


class RandomSelector(BaseRandomSelector):
    def __call__(self, dataset_pool, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        idx, idx_remain = random_choice(len(dataset_pool), self.batch_size)
        return idx, [], idx_remain

    @property
    def info(self) -> str:
        return f"RandomSelector(batch_size={self.batch_size})"


class ClusterRandomSelector(BaseClusterSelector):
    def __call__(self, dataset_pool, kernel: Callable, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        if self.batch_size < len(dataset_pool):
            idx_cluster = random_choice(len(dataset_pool), self.cluster_size)[0]
            idx = idx_cluster[np.array(self.get_idx_cluster(dataset_pool, kernel, idx_cluster))].tolist()

            mask = np.isin(range(len(dataset_pool)), idx)
            idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
            return idx, [], idx_remain
        else:
            return list(range(len(dataset_pool))), [], []

    @property
    def info(self) -> str:
        return f"ClusterRandomSelector(batch_size={self.batch_size}, cluster_size={self.cluster_size})"


class ExplorativeSelector(BaseRandomSelector):
    def __call__(self, model, dataset_pool, stop_cutoff: float = None, confidence_cutoff: float = None, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        if self.batch_size <= len(dataset_pool):
            y_std = model.predict_uncertainty(dataset_pool)
            idx = get_topn_idx(y_std, n=self.batch_size, target="max", cutoff=stop_cutoff)

            acquisition = y_std[np.array(idx)].tolist()

            idx_confident = np.where(y_std < confidence_cutoff)[0].tolist() if confidence_cutoff is not None else []
            mask = np.isin(range(len(dataset_pool)), idx + idx_confident)
            idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
            return idx, acquisition, idx_remain
        else:
            return list(range(len(dataset_pool))), [], []

    @property
    def info(self) -> str:
        return f"ExplorativeSelector(batch_size={self.batch_size})"


class ClusterExplorativeSelector(BaseClusterSelector):
    def __call__(self, model, dataset_pool, kernel: Callable, stop_cutoff: float = None, confidence_cutoff: float = None, 
                 **kwargs) -> Tuple[List[int], List[float], List[int]]:
        if self.batch_size <= len(dataset_pool):
            y_std = model.predict_uncertainty(dataset_pool)
            idx_cluster = get_topn_idx(y_std, n=self.cluster_size, target="max", cutoff=stop_cutoff)
            idx = np.array(idx_cluster)[np.array(self.get_idx_cluster(dataset_pool, kernel, idx_cluster))].tolist()

            acquisition = y_std[np.array(idx)].tolist()

            idx_confident = np.where(y_std < confidence_cutoff)[0].tolist() if confidence_cutoff is not None else []
            mask = np.isin(range(len(dataset_pool)), idx + idx_confident)
            idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
            return idx, acquisition, idx_remain
        else:
            return list(range(len(dataset_pool))), [], []

    @property
    def info(self) -> str:
        return f"ClusterExplorativeSelector(batch_size={self.batch_size}, cluster_size={self.cluster_size})"


class PartialQueryExplorativeSelector(BasePartialQuerySelector, BaseRandomSelector):
    def __init__(self, query_size: int, batch_size: int = 1, seed: int = 0):
        super().__init__(batch_size=batch_size, seed=seed)
        self.query_size = query_size
        assert batch_size <= query_size, "batch_size should be smaller than query_size."

    def __call__(self, model, dataset_pool, stop_cutoff: float = None, confidence_cutoff: float = None, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        if self.batch_size <= len(dataset_pool):
            data_query, idx_query = self.get_partial_data(dataset_pool, self.query_size)
            y_std = model.predict_uncertainty(data_query)
            idx_ = get_topn_idx(y_std, n=self.batch_size, target="max", cutoff=stop_cutoff)
            idx = idx_query[np.array(idx_)].tolist()

            acquisition = y_std[idx_].tolist()

            idx_confident_ = np.where(y_std < confidence_cutoff)[0].tolist() if confidence_cutoff is not None else []
            idx_confident = idx_query[np.array(idx_confident_)].tolist()
            mask = np.isin(range(len(dataset_pool)), idx + idx_confident)
            idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
            return idx, acquisition, idx_remain
        else:
            return list(range(len(dataset_pool))), [], []

    @property
    def info(self) -> str:
        return f"PartialQueryExplorativeSelector(batch_size={self.batch_size}, query_size={self.query_size})"


class PartialQueryClusterExplorativeSelector(BaseClusterSelector, BasePartialQuerySelector):
    def __init__(self, query_size: int, batch_size: int = 1, cluster_size: int = None, seed: int = 0):
        super().__init__(batch_size=batch_size, cluster_size=cluster_size, seed=seed)
        self.query_size = query_size
        assert self.cluster_size <= query_size, "cluster_size should be smaller than query_size."

    def __call__(self, model, dataset_pool, kernel: Callable, stop_cutoff: float = None, confidence_cutoff: float = None, 
                 **kwargs) -> Tuple[List[int], List[float], List[int]]:
        if self.batch_size <= len(dataset_pool):
            data_query, idx_query = self.get_partial_data(dataset_pool, self.query_size)
            y_std = model.predict_uncertainty(data_query)
            idx_cluster = get_topn_idx(y_std, n=self.cluster_size, target="max", cutoff=stop_cutoff)
            idx_ = np.array(idx_cluster)[np.array(self.get_idx_cluster(dataset_pool, kernel, idx_cluster))].tolist()
            idx = np.array(idx_query)[np.array(idx_)].tolist()

            acquisition = y_std[np.array(idx_)].tolist()

            idx_confident_ = np.where(y_std < confidence_cutoff)[0].tolist() if confidence_cutoff is not None else []
            idx_confident = idx_query[np.array(idx_confident_)].tolist()
            mask = np.isin(range(len(dataset_pool)), idx + idx_confident)
            idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
            return idx, acquisition, idx_remain
        else:
            return list(range(len(dataset_pool))), [], []

    @property
    def info(self) -> str:
        return (f"PartialQueryClusterExplorativeSelector(batch_size={self.batch_size}, "
                f"cluster_size={self.cluster_size}, query_size={self.query_size})")


class ExploitiveSelector(BaseRandomSelector):
    def __init__(self, target, batch_size: int = 1, seed: int = 0):
        super().__init__(batch_size=batch_size, seed=seed)
        self.target = target

    def __call__(self, model, dataset_pool, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        y_pred = np.asarray(model.predict_value(dataset_pool)).ravel()
        idx = get_topn_idx(y_pred, n=self.batch_size, target=self.target)

        acquisition = y_pred[np.array(idx)].tolist()

        mask = np.isin(range(len(dataset_pool)), idx)
        idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"ExploitiveSelector(batch_size={self.batch_size}, target={self.target})"


class ClusterExploitiveSelector(BaseClusterSelector):
    def __init__(self, target, batch_size: int = 1, cluster_size: int = None, seed: int = 0):
        super().__init__(batch_size=batch_size, cluster_size=cluster_size, seed=seed)
        self.target = target

    def __call__(self, model, dataset_pool, kernel: Callable, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        y_pred = np.asarray(model.predict_value(dataset_pool)).ravel()
        idx_cluster = get_topn_idx(y_pred, n=self.cluster_size, target=self.target)
        idx = np.array(idx_cluster)[np.array(self.get_idx_cluster(dataset_pool, kernel, idx_cluster))].tolist()

        acquisition = y_pred[np.array(idx)].tolist()

        mask = np.isin(range(len(dataset_pool)), idx)
        idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return (f"ClusterExploitiveSelector(batch_size={self.batch_size}, "
                f"cluster_size={self.cluster_size}, target={self.target})")


class PartialQueryExploitiveSelector(BasePartialQuerySelector, BaseRandomSelector):
    def __init__(self, target, query_size: int, batch_size: int = 1, seed: int = 0):
        super().__init__(batch_size=batch_size, seed=seed)
        self.target = target
        self.query_size = query_size
        assert batch_size <= query_size, "batch_size should be smaller than query_size."

    def __call__(self, model, dataset_pool, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        data_query, idx_query = self.get_partial_data(dataset_pool, self.query_size)
        y_pred = np.asarray(model.predict_value(data_query)).ravel()
        idx_ = get_topn_idx(y_pred, n=self.batch_size, target=self.target)
        idx = idx_query[np.array(idx_)].tolist()

        acquisition = y_pred[idx_].tolist()

        mask = np.isin(range(len(dataset_pool)), idx)
        idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return f"PartialQueryExploitiveSelector(batch_size={self.batch_size}, query_size={self.query_size}, target={self.target})"


class PartialQueryClusterExploitiveSelector(BaseClusterSelector, BasePartialQuerySelector):
    def __init__(self, target, query_size: int, batch_size: int = 1, cluster_size: int = None, seed: int = 0):
        super().__init__(batch_size=batch_size, cluster_size=cluster_size, seed=seed)
        self.target = target
        self.query_size = query_size
        assert self.cluster_size <= query_size, "cluster_size should be smaller than query_size."
    
    def __call__(self, model, dataset_pool, kernel: Callable, **kwargs) -> Tuple[List[int], List[float], List[int]]:
        data_query, idx_query = self.get_partial_data(dataset_pool, self.query_size)
        y_pred = np.asarray(model.predict_value(data_query)).ravel()
        idx_cluster = get_topn_idx(y_pred, n=self.cluster_size, target=self.target)
        idx_ = np.array(idx_cluster)[np.array(self.get_idx_cluster(dataset_pool, kernel, idx_cluster))].tolist()
        idx = np.array(idx_query)[np.array(idx_)].tolist()

        acquisition = y_pred[idx_].tolist()

        mask = np.isin(range(len(dataset_pool)), idx)
        idx_remain = np.arange(len(dataset_pool))[~mask].tolist()
        return idx, acquisition, idx_remain

    @property
    def info(self) -> str:
        return (f"PartialQueryClusterExploitiveSelector(batch_size={self.batch_size}, "
                f"cluster_size={self.cluster_size}, query_size={self.query_size}, target={self.target})")


class ProbabilityImprovementSelector(BaseRandomSelector):
    pass


class ExpectedImprovementSelector(BaseRandomSelector):
    pass


class UpperConfidenceBoundSelector(BaseRandomSelector):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from sklearn.ensemble import GradientBoostingClassifier as GBC
from sklearn.ensemble import GradientBoostingRegressor as GBR
from molalkit.models.base import BaseSklearnModel


class GradientBoostingClassifier(GBC, BaseSklearnModel):
    def fit_molalkit(self, train_data, iteration: int = 0):
        return self.fit_molalkit_(train_data, self)

    def predict_uncertainty(self, pred_data):
        return self.predict_uncertainty_c(pred_data, self)

    def predict_value(self, pred_data):
        return self.predict_value_c(pred_data, self)


class GradientBoostingRegressor(GBR, BaseSklearnModel):
    def fit_molalkit(self, train_data, iteration: int = 0):
        return self.fit_molalkit_(train_data, self)

    def predict_uncertainty(self, pred_data):
        return self.predict_uncertainty_r(pred_data, self)

    def predict_value(self, pred_data):
        return self.predict_value_r(pred_data, self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict, List, Optional, Literal, Tuple, Callable
from tap import Tap
import os
import shutil
import json
import math
import copy
import pandas as pd
import numpy as np
from mgktools.data.split import data_split_index
from mgktools.evaluators.metric import Metric
from mgktools.features_mol.features_generators import FeaturesGenerator
from molalkit.active_learning.selector import (
    RandomSelector, ClusterRandomSelector, 
    ExplorativeSelector, ClusterExplorativeSelector,
    PartialQueryExplorativeSelector, PartialQueryClusterExplorativeSelector,
    ExploitiveSelector, ClusterExploitiveSelector,
    PartialQueryExploitiveSelector, PartialQueryClusterExploitiveSelector,
)
from molalkit.active_learning.forgetter import (
    FirstForgetter, RandomForgetter, 
    MinOOBUncertaintyForgetter, MaxOOBUncertaintyForgetter,
    MinOOBErrorForgetter, MaxOOBErrorForgetter, 
    MinLOOUncertaintyForgetter, MaxLOOUncertaintyForgetter,
    MinLOOErrorForgetter, MaxLOOErrorForgetter
)
from molalkit.active_learning.utils import get_topn_idx
from molalkit.data.datasets import DATA_DIR
from molalkit.data.utils import get_data, get_subset_from_uidx
from molalkit.exe.logging import create_logger
from molalkit.exe.utils import (
    read_df,
    get_features_generators_from_config,
    get_kernel_from_config,
    get_model_from_config,
    add_error_rate_to_labels,
    apply_error_rate_to_id2datapoint,
)
from molalkit.models.base import BaseSklearnModel
from molalkit.models.configs import MODEL_DIR


class CommonArgs(Tap):
    save_dir: str
    """the output directory."""
    n_jobs: int = 1
    """the cpu numbers used for parallel computing."""
    logger_name: str = "alb_output"
    """the prefix of the output logger file: verbose.log and quite.log"""
    verbose: int = 1
    """the level of verbosity. 0, 1, 2"""
    seed: int = 0
    """random seed."""

    def process_args(self) -> None:
        os.makedirs(self.save_dir, exist_ok=True)
        self.logger = create_logger(self.logger_name, save_dir=self.save_dir, verbose=self.verbose)
        np.random.seed(self.seed)


class DatasetArgs(CommonArgs):
    data_public = None
    """Use public data sets."""
    data_path: str = None
    """the Path of input data CSV file."""
    data_path_training: str = None
    """the Path of input data CSV file for training set."""
    data_path_pool: str = None
    """the Path of input data CSV file for pool set."""
    data_path_val: str = None
    """the Path of input data CSV file for validation set."""
    smiles_columns: List[str] = None
    """
    For pure compounds.
    Name of the columns containing single SMILES or InChI string.
    """
    targets_columns: List[str] = None
    """
    Name of the columns containing target values.
    """
    features_columns: List[str] = None
    """
    Name of the columns containing additional features_mol such as temperature, 
    pressuer.
    """
    task_type: Literal["regression", "binary", "multiclass"] = None
    """
    Type of task.
    """
    split_type: Literal["random", "scaffold_random", "scaffold_order"] = None
    """Method of splitting the data into active learning/validation."""
    split_sizes: List[float] = None
    """Split proportions for active learning/validation sets."""
    full_val: bool = False
    """validate the performance of active learning on the full dataset."""
    init_size: int = 2
    """number of samples as the initial."""
    error_rate: float = None
    """the percent of the training set that will be affected by error (0-1), valid only for binary classification."""

    def get_train_pool_split_index(self, df_al: pd.DataFrame) -> Tuple[List[int], List[int]]:
        if self.init_size >= len(df_al):
            train_index, pool_index = list(range(len(df_al))), []
        else:
            if self.task_type == "regression":
                train_index, pool_index = data_split_index(
                    n_samples=len(df_al),
                    mols=df_al[self.smiles_columns[0]] if self.smiles_columns is not None else None,
                    split_type="random",
                    sizes=[self.init_size / len(df_al), 1 - self.init_size / len(df_al)],
                    seed=self.seed)
            else:
                train_index, pool_index = data_split_index(
                    n_samples=len(df_al),
                    mols=df_al[self.smiles_columns[0]] if self.smiles_columns is not None else None,
                    targets=df_al[self.targets_columns[0]],
                    split_type="init_al",
                    n_samples_per_class=1,
                    seed=self.seed)
                # randomly select self.init_size - 2 samples from the pool set to be the training set
                if self.init_size > 2:
                    train_index.extend(np.random.choice(pool_index, self.init_size - 2, replace=False))
                    pool_index = [i for i in pool_index if i not in train_index]
        return train_index, pool_index

    def process_args(self) -> None:
        super().process_args()
        if self.data_public == "freesolv" or self.data_public == "test_regression":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["freesolv"]
            self.task_type = "regression"
        elif self.data_public == "delaney":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["logSolubility"]
            self.task_type = "regression"
        elif self.data_public == "lipo":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["lipo"]
            self.task_type = "regression"
        elif self.data_public == "pdbbind_refined":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["-logKd/Ki"]
            self.task_type = "regression"
        elif self.data_public == "pdbbind_full":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["-logKd/Ki"]
            self.task_type = "regression"
        elif self.data_public in ["ld50_zhu", "caco2_wang", "solubility_aqsoldb", "ppbr_az", "vdss_lombardo",
                                  "Half_Life_Obach", "Clearance_Hepatocyte_AZ"]:
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["Drug"]
            self.targets_columns = ["Y"]
            self.task_type = "regression"
        elif self.data_public == "bbbp":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["p_np"]
            self.task_type = "binary"
        elif self.data_public == "bace":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["mol"]
            self.targets_columns = ["Class"]
            self.task_type = "binary"
        elif self.data_public == "hiv":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["smiles"]
            self.targets_columns = ["HIV_active"]
            self.task_type = "binary"
        elif self.data_public in ["ames", "carcinogens_lagunin", "dili", "herg", "skin", "hia_hou", "pgp_broccatelli",
                                  "bioavailability_ma", "clintox", "bbb_martins", "CYP1A2_Veith",
                                  "CYP2C9_Substrate_CarbonMangels", "CYP2C9_Veith", "CYP2C19_Veith",
                                  "CYP2D6_Substrate_CarbonMangels", "CYP2D6_Veith", "CYP3A4_Veith",
                                  "CYP3A4_Substrate_CarbonMangels"]:
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["Drug"]
            self.targets_columns = ["Y"]
            self.task_type = "binary"
        elif self.data_public == "human_liver_microsome_stability":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["SMILES"]
            self.targets_columns = ["LOG HLM_CLint (mL/min/kg)"]
            self.task_type = "regression"
        elif self.data_public == "rat_liver_microsome_stability":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["SMILES"]
            self.targets_columns = ["LOG RLM_CLint (mL/min/kg)"]
            self.task_type = "regression"
        elif self.data_public == "MDRR1-MDCK_efflux_ratio":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["SMILES"]
            self.targets_columns = ["LOG MDR1-MDCK ER (B-A/A-B)"]
            self.task_type = "regression"
        elif self.data_public == "aqueous_solubility":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["SMILES"]
            self.targets_columns = ["LOG SOLUBILITY PH 6.8 (ug/mL)"]
            self.task_type = "regression"
        elif self.data_public == "human_plasma_protein_binding":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["SMILES"]
            self.targets_columns = ["LOG PLASMA PROTEIN BINDING (HUMAN) (% unbound)"]
        elif self.data_public == "rat_plasma_protein_binding":
            self.data_path = os.path.join(DATA_DIR, "%s.csv" % self.data_public)
            self.smiles_columns = ["SMILES"]
            self.targets_columns = ["LOG PLASMA PROTEIN BINDING (RAT) (% unbound)"]

        if self.split_type is not None and self.split_type.startswith("scaffold"):
            assert len(self.smiles_columns) == 1, f"Scaffold split is valid only for single SMILES column, but {len(self.smiles_columns)} are provided."

        assert len(self.targets_columns) == 1, "multi-task active learning is not implemented yet."

        if self.data_path is not None:
            # All data comes from the same file.
            assert self.data_path_val is None and self.data_path_training is None and self.data_path_pool is None
            df = read_df(self.data_path, self.task_type, self.targets_columns)
            if "uidx" not in df:
                df["uidx"] = range(len(df))
            if self.full_val:
                # use the full dataset as validation set
                assert self.split_type is None
                assert self.split_sizes is None
                assert self.error_rate is None
                df.to_csv("%s/val.csv" % self.save_dir, index=False)
                df_al = df
            else:
                # split the dataset into active learning and validation sets
                al_index, val_index = data_split_index(
                    n_samples=len(df),
                    mols=df[self.smiles_columns[0]] if self.smiles_columns is not None else None,
                    # targets=df[self.targets_columns[0]],
                    split_type=self.split_type,
                    sizes=self.split_sizes,
                    seed=self.seed,
                    logger=self.logger)
                df[df.index.isin(val_index)].to_csv("%s/val.csv" % self.save_dir, index=False)
                df_al = df[df.index.isin(al_index)].copy()
                if self.error_rate is not None:
                    assert self.task_type == "binary", "error_rate is valid only for binary classification."
                    # randomly select a portion of the training set to be affected by error
                    add_error_rate_to_labels(df_al, self.error_rate, self.targets_columns[0])
            # split the active learning set into training and pool sets
            train_index, pool_index = self.get_train_pool_split_index(df_al)
            df_al.iloc[train_index].to_csv("%s/train_init.csv" % self.save_dir, index=False)
            df_al.iloc[pool_index].to_csv("%s/pool_init.csv" % self.save_dir, index=False)
        else:
            # data comes from 3 different files.
            assert self.data_path_training is not None, "please provide input data"
            if self.data_path_pool is None:
                df_al = read_df(self.data_path_training, self.task_type, self.targets_columns)
                train_index, pool_index = self.get_train_pool_split_index(df_al)
                df_al.iloc[train_index].to_csv("%s/train_init.csv" % self.save_dir, index=False)
                df_al.iloc[pool_index].to_csv("%s/pool_init.csv" % self.save_dir, index=False)
            else:
                df_train = read_df(self.data_path_training, self.task_type, self.targets_columns)
                df_train.to_csv("%s/train_init.csv" % self.save_dir, index=False)
                df_pool = read_df(self.data_path_pool, self.task_type, self.targets_columns)
                df_pool.to_csv("%s/pool_init.csv" % self.save_dir, index=False)
            if self.data_path_val is None:
                df_train.sample(0).to_csv("%s/val.csv" % self.save_dir, index=False)
                df = pd.concat([pd.read_csv(f) for f in ["%s/train_init.csv" % self.save_dir,
                                                         "%s/pool_init.csv" % self.save_dir]])
            else:
                df_val = read_df(self.data_path_val, self.task_type, self.targets_columns)
                df_val.to_csv("%s/val.csv" % self.save_dir, index=False)
                df = pd.concat([pd.read_csv(f) for f in ["%s/train_init.csv" % self.save_dir,
                                                         "%s/pool_init.csv" % self.save_dir,
                                                         "%s/val.csv" % self.save_dir]])
            if "uidx" not in df:
                df["uidx"] = range(len(df))
                df_train = pd.read_csv("%s/train_init.csv" % self.save_dir)
                df_train["uidx"] = range(len(df_train))
                df_pool = pd.read_csv("%s/pool_init.csv" % self.save_dir)
                df_pool["uidx"] = range(len(df_pool))
                df_pool["uidx"] += len(df_train)
                df_val = pd.read_csv("%s/val.csv" % self.save_dir)
                df_val["uidx"] = range(len(df_val))
                df_val["uidx"] += len(df_train) + len(df_pool)
                if self.error_rate is not None:
                    add_error_rate_to_labels(df_train, self.error_rate, self.targets_columns[0])
                    add_error_rate_to_labels(df_pool, self.error_rate, self.targets_columns[0])
                df_train.to_csv("%s/train_init.csv" % self.save_dir, index=False)
                df_pool.to_csv("%s/pool_init.csv" % self.save_dir, index=False)
                df_val.to_csv("%s/val.csv" % self.save_dir, index=False)
        df.to_csv("%s/full.csv" % self.save_dir, index=False)


class ModelArgs(Tap):
    model_configs: List[str]
    """ A list of config files contain all information of the machine learning model for performance evaluation.
        The first one will be used as the selector model, and the rest will be used as evaluators."""

    @property
    def model_configs_dict(self) -> List[Dict]:
        return [json.loads(open(self.find_model_config_file(m)).read()) for m in self.model_configs]

    @staticmethod
    def find_model_config_file(config_name):
        if os.path.exists(config_name):
            # try to find the config file in the current directory
            return config_name
        else:
            # try to find the config file in MolALKit"s model directory
            stored_config_name = f"{MODEL_DIR}/{config_name}"
            assert os.path.exists(stored_config_name), f"{config_name} not found."
            return stored_config_name


class DatasetModelArgs(DatasetArgs, ModelArgs):
    @property
    def features_generators(self) -> Optional[List[List[FeaturesGenerator]]]:
        return [get_features_generators_from_config(model_config) for model_config in self.model_configs_dict]

    @property
    def kernels(self) -> List[Callable]:
        if not hasattr(self, "_kernels"):
            self._kernels = [get_kernel_from_config(
                model_config=model_config,
                dataset=self.datasets_full[i],
                kernel_pkl_path="%s/kernels_%d.pkl" % (self.save_dir, i),
            ) for i, model_config in enumerate(self.model_configs_dict)]
        return self._kernels

    @property
    def models(self) -> List[BaseSklearnModel]:
        if not hasattr(self, "_models"):
            self._models = [get_model_from_config(
                model_config=model_config,
                dataset=self.datasets_full[i],
                task_type=self.task_type,
                save_dir=self.save_dir,
                data_path="%s/full.csv" % self.save_dir,
                smiles_columns=self.smiles_columns,
                targets_columns=self.targets_columns,
                features_generators=self.features_generators[i],
                kernel=self.kernels[i],
                n_jobs=self.n_jobs,
                seed=self.seed,
                logger=self.logger
            ) for i, model_config in enumerate(self.model_configs_dict)]
        return self._models

    @property
    def datasets_full(self) -> List:
        if not hasattr(self, "_datasets_full"):
            self._datasets_full = [get_data(
                data_format=model_config["data_format"],
                save_dir=self.save_dir,
                file_name="full.csv",
                smiles_columns=self.smiles_columns,
                targets_columns=self.targets_columns,
                features_columns=self.features_columns,
                features_generators=self.features_generators[i],
                features_combination=model_config.get("features_combination"),
                graph_kernel_type=model_config.get("graph_kernel_type") or "no",
                n_jobs=self.n_jobs) for i, model_config in enumerate(self.model_configs_dict)]
        return self._datasets_full

    @property
    def datasets_empty(self) -> List:
        if not hasattr(self, "_datasets_empty"):
            self._datasets_empty = []
            for dataset in self.datasets_full:
                dataset_empty = copy.copy(dataset)
                dataset_empty.data = []
                self._datasets_empty.append(dataset_empty)
        return self._datasets_empty.copy()

    @property
    def id2datapoints(self) -> List[Dict]:
        if not hasattr(self, "_id2datapoints"):
            self._id2datapoints = [{data.uidx: data for data in dataset} for dataset in self.datasets_full]
            if self.error_rate:
                for id2datapoint in self._id2datapoints:
                    df = pd.read_csv("%s/train_init.csv" % self.save_dir)
                    apply_error_rate_to_id2datapoint(id2datapoint, df, self.targets_columns[0])
                    df = pd.read_csv("%s/pool_init.csv" % self.save_dir)
                    apply_error_rate_to_id2datapoint(id2datapoint, df, self.targets_columns[0])
        return self._id2datapoints

    @property
    def datasets_train(self):
        if not hasattr(self, "_datasets_train"):
            df = pd.read_csv("%s/train_init.csv" % self.save_dir)
            self._datasets_train = [get_subset_from_uidx(dataset, id2datapoint, df["uidx"].tolist()) 
                                    for dataset, id2datapoint in zip(self.datasets_empty, self.id2datapoints)]
        return self._datasets_train
    
    @property
    def datasets_pool(self):
        if not hasattr(self, "_datasets_pool"):
            df = pd.read_csv("%s/pool_init.csv" % self.save_dir)
            self._datasets_pool = [get_subset_from_uidx(dataset, id2datapoint, df["uidx"].tolist()) 
                                   for dataset, id2datapoint in zip(self.datasets_empty, self.id2datapoints)]
        return self._datasets_pool

    @property
    def datasets_val(self):
        if not hasattr(self, "_datasets_val"):
            df = pd.read_csv("%s/val.csv" % self.save_dir)
            self._datasets_val = [get_subset_from_uidx(dataset, id2datapoint, df["uidx"].tolist()) 
                                  for dataset, id2datapoint in zip(self.datasets_empty, self.id2datapoints)]
        return self._datasets_val
    
    def process_args(self) -> None:
        super().process_args()


class SelectorArgs(Tap):
    select_method: Literal["random", "explorative", "exploitive"] = None
    """the method to select the next batch of samples."""
    s_batch_size: int = 1
    """number of samples to select in each iteration."""
    s_batch_mode: Literal["naive", "cluster"] = "naive"
    """the method to select the next batch of samples."""
    s_cluster_size: int = None
    """number of samples in each cluster for cluster batch selection. (default = 20 * batch_size)"""
    s_query_size: int = None
    """number of samples to query in each active learning iteration. (default=None means query all samples in the 
    pool set)"""
    s_exploitive_target: str = None
    """the target value for exploitive active learning."""
    seed: int = 0
    """random seed."""

    @property
    def selector(self):
        if not hasattr(self, "_selector"):
            if self.s_exploitive_target is not None and self.s_exploitive_target not in ["min", "max"]:
                self.s_exploitive_target = float(self.s_exploitive_target)
            if self.select_method == "random":
                if self.s_batch_mode == "cluster":
                    self._selector = ClusterRandomSelector(batch_size=self.s_batch_size,
                                                           cluster_size=self.s_cluster_size,
                                                           seed=self.seed)
                else:
                    self._selector = RandomSelector(batch_size=self.s_batch_size, seed=self.seed)
            elif self.select_method == "explorative":
                if self.s_batch_mode == "cluster":
                    if self.s_query_size is not None:
                        self._selector = PartialQueryClusterExplorativeSelector(
                            batch_size=self.s_batch_size,
                            cluster_size=self.s_cluster_size,
                            query_size=self.s_query_size,
                            seed=self.seed)
                    else:
                        self._selector = ClusterExplorativeSelector(batch_size=self.s_batch_size,
                                                                    cluster_size=self.s_cluster_size,
                                                                    seed=self.seed)
                else:
                    if self.s_query_size is not None:
                        self._selector = PartialQueryExplorativeSelector(batch_size=self.s_batch_size,
                                                                         query_size=self.s_query_size,
                                                                         seed=self.seed)
                    else:
                        self._selector = ExplorativeSelector(batch_size=self.s_batch_size, seed=self.seed)
            elif self.select_method == "exploitive":
                if self.s_batch_mode == "cluster":
                    if self.s_query_size is not None:
                        self._selector = PartialQueryClusterExploitiveSelector(
                            batch_size=self.s_batch_size,
                            cluster_size=self.s_cluster_size,
                            query_size=self.s_query_size,
                            target=self.s_exploitive_target,
                            seed=self.seed)
                    else:
                        self._selector = ClusterExploitiveSelector(batch_size=self.s_batch_size,
                                                                   cluster_size=self.s_cluster_size,
                                                                   target=self.s_exploitive_target,
                                                                   seed=self.seed)
                else:
                    if self.s_query_size is not None:
                        self._selector = PartialQueryExploitiveSelector(batch_size=self.s_batch_size, 
                                                                        query_size=self.s_query_size,
                                                                        target=self.s_exploitive_target,
                                                                        seed=self.seed)
                    else:
                        self._selector = ExploitiveSelector(batch_size=self.s_batch_size, 
                                                            target=self.s_exploitive_target, 
                                                            seed=self.seed)
            elif self.select_method is None:
                return None
            else:
                raise ValueError(f"Unknown selection method {self.select_method}")
        return self._selector


class ForgetterArgs(Tap):
    forget_method: Literal["first", "random", 
                           "min_oob_uncertainty", "max_oob_uncertainty", 
                           "min_oob_error", "max_oob_error",
                           "min_loo_uncertainty", "max_loo_uncertainty",
                           "min_loo_error", "max_loo_error"] = None
    """the forget method."""
    f_batch_size: int = 1
    """number of samples to forget in each iteration."""
    f_min_train_size: int = None
    """Forget method activate only when the number of samples in the training set larger than this number."""
    seed: int = 0
    """random seed."""

    @property
    def forgetter(self):
        if not hasattr(self, "_forgetter"):
            if self.forget_method == "first":
                self._forgetter = FirstForgetter(batch_size=self.f_batch_size)
            elif self.forget_method == "random":
                self._forgetter = RandomForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "min_oob_uncertainty":
                self._forgetter = MinOOBUncertaintyForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "max_oob_uncertainty":
                self._forgetter = MaxOOBUncertaintyForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "min_oob_error":
                self._forgetter = MinOOBErrorForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "max_oob_error":
                self._forgetter = MaxOOBErrorForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "min_loo_uncertainty":
                self._forgetter = MinLOOUncertaintyForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "max_loo_uncertainty":
                self._forgetter = MaxLOOUncertaintyForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "min_loo_error":
                self._forgetter = MinLOOErrorForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method == "max_loo_error":
                self._forgetter = MaxLOOErrorForgetter(batch_size=self.f_batch_size, seed=self.seed)
            elif self.forget_method is None:
                return None
            else:
                raise ValueError(f"Unknown forget method {self.forget_method}")
        return self._forgetter


class EvaluationArgs(Tap):
    metrics: List[Metric] = None
    """the metrics to evaluate model performance."""
    evaluate_stride: int = None
    """evaluate model performance on the validation set when the size of the training set is an integer multiple of the 
    evaluation stride."""
    top_k: float = None
    """ For regression tasks, the ratio of top molecules are considered.
        For binary classification tasks, the molecules with label 1 will considered as top molecules.
    """
    detail: bool = False
    """output the details of each active learning iteration."""


class LearningArgs(DatasetModelArgs, SelectorArgs, ForgetterArgs, EvaluationArgs):
    n_select: int = 1
    """number of selecting steps in each select-forget loop."""
    n_forget: int = 0
    """number of forgetting steps in each select-forget loop."""
    max_iter: int = None
    """the maximum number of select-forget loops."""
    save_cpt_stride: int = None
    """save checkpoint file every no. steps of select-forget loops."""
    write_traj_stride: int = 1
    """write trajectory file every no. steps of select-forget loops."""
    load_checkpoint: bool = False
    """load checkpoint file and continue the active learning."""

    @property
    def top_uidx(self) -> Optional[List[int]]:
        if self.top_k is not None:
            top_uidx = []
            if self.task_type == "binary":
                for data in (self.datasets_train[0].data + self.datasets_pool[0].data):
                    if data.targets[0] == 1:
                        top_uidx.append(data.uidx)
            else:
                assert 0. < self.top_k < 1., "top_k must be in (0, 1)."
                n_top_k = math.ceil(self.top_k * (len(self.datasets_train[0]) + len(self.datasets_pool[0])))
                y_AL = self.datasets_train[0].y.ravel().tolist() + self.datasets_pool[0].y.ravel().tolist()
                top_k_index = get_topn_idx(y_AL, n_top_k, target=self.s_exploitive_target)
                for i, data in enumerate(self.datasets_train[0].data + self.datasets_pool[0].data):
                    if i in top_k_index:
                        top_uidx.append(data.uidx)
            return top_uidx
        else:
            return None
        
    def process_args(self) -> None:
        super().process_args()
        for i, model_config_dict in enumerate(self.model_configs_dict):
            if model_config_dict.get("graph_kernel_type") == "graph":
                self.datasets_full[i].unify_datatype()
        # check the input for exploitive active learning
        if self.select_method == "exploitive":
            # assert self.task_type == "regression", "exploitive active learning only support regression task."
            # assert self.top_k is not None, "top_k must be set for exploitive active learning."
            assert self.s_exploitive_target is not None, "s_exploitive_target must be set for exploitive active learning."

        if self.select_method is not None:
            assert self.n_select > 0, "n_select must be greater than 0."
        if self.forget_method is not None:
            assert self.n_forget > 0, "n_forget must be greater than 0."

        if self.max_iter is None:
            dn = self.n_select - self.n_forget
            if dn > 0:
                self.max_iter = len(self.datasets_pool[0]) // dn
            elif dn < 0:
                self.max_iter = len(self.datasets_train[0]) // (-dn) - 1
            else:
                self.max_iter = len(self.datasets_pool[0]) // self.n_select



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import List, Union, Literal
import pickle
import os
from logging import Logger
from sklearn.gaussian_process.kernels import RBF, DotProduct
from mgktools.data.data import Dataset
from mgktools.kernels.utils import get_kernel_config
from mgktools.kernels.PreComputed import calc_precomputed_kernel_config
from molalkit.exe.logging import EmptyLogger


def get_model(data_format: Literal["mgktools", "chemprop", "graphgps"],
              task_type: Literal["regression", "binary", "multiclass"],
              n_jobs: int = 8,
              seed: int = 0,
              # arguments for classical machine learning models
              model: Literal["random_forest", "naive_bayes", "logistic_regression", "gaussian_process_regression",
                             "gaussian_process_classification", "support_vector_machine", "adaboost", "xgboost", 
                             "decision_tree", "extra_trees", "MultinomialNB", "BernoulliNB", "GaussianNB", 
                             "LSTM", "GRU", "MolFormer"] = "random_forest",
              kernel=None,
              uncertainty_type: Literal["value", "uncertainty"] = None,
              alpha: Union[float, str] = 1e-8,
              C: float = 1.0,
              booster: Literal["gbtree", "gblinear", "dart"] = "gbtree",
              n_estimators: int = 100,
              max_depth: int = None,
              learning_rate: float = 0.1,
              tokenizer: Literal["SMILES", "SELFIES"] = None,
              smiles_full: List[str] = None,
              embedding_size: int = 64,
              # chemprop arguments
              save_dir: str = None,
              data_path: str = None,
              smiles_columns: List[str] = None,
              target_columns: List[str] = None,
              loss_function: Literal["mse", "bounded_mse", "binary_cross_entropy", "cross_entropy", "mcc", "sid",
                                     "wasserstein", "mve", "evidential", "dirichlet"] = None,
              multiclass_num_classes: int = 3,
              features_generator=None,
              no_features_scaling: bool = False,
              features_only: bool = False,
              features_size: int = 0,
              epochs: int = 30,
              depth: int = 3,
              hidden_size: int = 300,
              ffn_num_layers: int = 2,
              ffn_hidden_size: int = None,
              dropout: float = 0.0,
              batch_size: int = 50,
              ensemble_size: int = 1,
              number_of_molecules: int = 1,
              mpn_shared: bool = False,
              atom_messages: bool = False,
              undirected: bool = False,
              const_lr: bool = False,
              weight_decay: float = 0.0,
              cbp: bool = False,
              replacement_rate: float = 0.00001,
              maturity_threshold: int = 100,
              reinit_weights: Literal['xavier', 'kaiming', 'lecun', 'default'] = 'xavier',
              decay_rate: float = 0.99,
              class_balance: bool = False,
              checkpoint_dir: str = None,
              checkpoint_frzn: str = None,
              frzn_ffn_layers: int = 0,
              freeze_first_only: bool = False,
              mpn_path: str = None,
              freeze_mpn: bool = False,
              uncertainty_method: Literal["mve", "ensemble", "evidential_epistemic", "evidential_aleatoric",
                                          "evidential_total", "classification", "dropout", "spectra_roundrobin"] = None,
              uncertainty_dropout_p: float = 0.1,
              dropout_sampling_size: int = 10,
              continuous_fit: bool = False,
              # graphgps arguments
              cfg_path: str = None,
              # MolFormer arguments
              pretrained_path: str = None,
              n_head: int = 12,
              n_layer: int = 12,
              n_embd: int = 768,
              d_dropout: float = 0.1,
              num_feats: int = 32,
              logger: Logger = None):
    if alpha.__class__ == str:
        alpha = float(open(alpha).read())

    if data_format == "mgktools":
        if model == "random_forest":
            if task_type == "regression":
                from molalkit.models.random_forest.RandomForestRegressor import RFRegressor
                return RFRegressor(n_estimators=n_estimators, max_depth=max_depth, n_jobs=n_jobs, random_state=seed)
            else:
                from molalkit.models.random_forest.RandomForestClassifier import RFClassifier
                return RFClassifier(n_estimators=n_estimators, max_depth=max_depth, n_jobs=n_jobs, random_state=seed, oob_score=True)
        elif model == "MultinomialNB":
            assert task_type == "binary"
            from molalkit.models.naive_bayes.NaiveBayesClassifier import MultinomialNBClassifier
            return MultinomialNBClassifier()
        elif model == "BernoulliNB":
            assert task_type == "binary"
            from molalkit.models.naive_bayes.NaiveBayesClassifier import BernoulliNBClassifier
            return BernoulliNBClassifier()
        elif model == "GaussianNB":
            assert task_type == "binary"
            from molalkit.models.naive_bayes.NaiveBayesClassifier import GaussianNBClassifier
            return GaussianNBClassifier()
        elif model == "logistic_regression":
            assert task_type == "binary"
            from molalkit.models.logistic_regression.LogisticRegression import LogisticRegressor
            return LogisticRegressor(random_state=seed)
        elif model == "decision_tree":
            assert task_type == "binary"
            from molalkit.models.dt.dt import DecisionTreeClassifier
            return DecisionTreeClassifier(max_depth=max_depth, random_state=seed)
        elif model == "extra_trees":
            assert task_type == "binary"
            from molalkit.models.extra_trees.extra_trees import ExtraTreesClassifier
            return ExtraTreesClassifier(n_estimators=n_estimators, max_depth=max_depth, n_jobs=n_jobs, random_state=seed)
        elif model == "gaussian_process_regression":
            assert task_type in ["regression", "binary"]
            assert uncertainty_type is not None
            from molalkit.models.gaussian_process.GaussianProcessRegressor import GPRegressor
            return GPRegressor(kernel=kernel, alpha=alpha, optimizer=None, uncertainty_type=uncertainty_type)
        elif model == "gaussian_process_classification":
            assert task_type == "binary"
            from molalkit.models.gaussian_process.GaussianProcessClassifier import GPClassifier
            return GPClassifier(kernel=kernel, optimizer=None)
        elif model == "support_vector_machine":
            if task_type == "regression":
                from molalkit.models.support_vector.SupportVectorRegressor import SVRegressor
                return SVRegressor(kernel=kernel, C=C)
            else:
                from molalkit.models.support_vector.SupportVectorClassifier import SVClassifier
                return SVClassifier(kernel=kernel, C=C, probability=True)
        elif model == "adaboost":
            if task_type == "regression":
                from molalkit.models.adaboost.AdaBoostRegressor import AdaBoostRegressor
                return AdaBoostRegressor(random_state=seed)
            else:
                from molalkit.models.adaboost.AdaBoostClassifier import AdaBoostClassifier
                return AdaBoostClassifier(random_state=seed)
        elif model == "xgboost":
            if task_type == "regression":
                from molalkit.models.xgboost.XGBRegressor import XGBRegressor
                return XGBRegressor(booster=booster, n_estimators=n_estimators, max_depth=max_depth, learning_rate=learning_rate, n_jobs=n_jobs, random_state=seed)
            else:
                from molalkit.models.xgboost.XGBClassifier import XGBClassifier
                return XGBClassifier(booster=booster, n_estimators=n_estimators, max_depth=max_depth, learning_rate=learning_rate, n_jobs=n_jobs, random_state=seed)
        elif model == "gradient_boosting":
            if task_type == "regression":
                from molalkit.models.gradient_boosting.gradient_boosting import GradientBoostingRegressor
                return GradientBoostingRegressor(n_estimators=n_estimators, max_depth=max_depth, learning_rate=learning_rate, random_state=seed)
            else:
                from molalkit.models.gradient_boosting.gradient_boosting import GradientBoostingClassifier
                return GradientBoostingClassifier(n_estimators=n_estimators, max_depth=max_depth, learning_rate=learning_rate, random_state=seed)
        elif model in ["LSTM", "GRU"]:
            from molalkit.models.rnn.rnn import RNN
            from molalkit.models.rnn.tokenizer import SMILESTokenizer, SELFIESTokenizer
            if tokenizer == "SMILES":
                tokenizer_ = SMILESTokenizer()
            elif tokenizer == "SELFIES":
                tokenizer_ = SELFIESTokenizer()
            else:
                raise ValueError(f"unknown tokenizer: {tokenizer}")
            tokenizer_.create_vocabulary(smiles_full)
            return RNN(tokenizer=tokenizer_,
                       task_type=task_type,
                       rnn_type=model,
                       embedding_size=embedding_size,
                       depth=depth,
                       hidden_size=hidden_size,
                       dropout=dropout,
                       epochs=epochs,
                       ffn_num_layers=ffn_num_layers,
                       batch_size=batch_size)
        elif model == "MolFormer":
            from molformer.molformer import MolFormer
            return MolFormer(save_dir=save_dir, task_type=task_type, pretrained_path=pretrained_path,
                             num_tasks=len(target_columns),
                             n_head=n_head, n_layer=n_layer, n_embd=n_embd, d_dropout=d_dropout,
                             dropout=dropout, learning_rate=learning_rate, 
                             num_feats=num_feats, ensemble_size=ensemble_size, epochs=epochs,
                             batch_size=batch_size, num_workers=n_jobs, seed=seed)
        else:
            raise ValueError(f"unknown model: {model}")
    elif data_format == "chemprop":
        from chemprop.models.mpnn import MPNN
        return MPNN(save_dir=save_dir,
                    data_path=data_path,
                    smiles_columns=smiles_columns,
                    target_columns=target_columns,
                    dataset_type="classification" if task_type == "binary" else task_type,
                    loss_function=loss_function,
                    multiclass_num_classes=multiclass_num_classes,
                    features_generator=features_generator,
                    no_features_scaling=no_features_scaling,
                    features_only=features_only,
                    features_size=features_size,
                    epochs=epochs,
                    depth=depth,
                    hidden_size=hidden_size,
                    ffn_num_layers=ffn_num_layers,
                    ffn_hidden_size=ffn_hidden_size,
                    dropout=dropout,
                    batch_size=batch_size,
                    ensemble_size=ensemble_size,
                    number_of_molecules=number_of_molecules,
                    mpn_shared=mpn_shared,
                    atom_messages=atom_messages,
                    undirected=undirected,
                    const_lr=const_lr,
                    weight_decay=weight_decay,
                    cbp=cbp,
                    replacement_rate=replacement_rate,
                    maturity_threshold=maturity_threshold,
                    reinit_weights=reinit_weights,
                    decay_rate=decay_rate,
                    class_balance=class_balance,
                    checkpoint_dir=checkpoint_dir,
                    checkpoint_frzn=checkpoint_frzn,
                    frzn_ffn_layers=frzn_ffn_layers,
                    freeze_first_only=freeze_first_only,
                    mpn_path=mpn_path,
                    freeze_mpn=freeze_mpn,
                    uncertainty_method=uncertainty_method,
                    uncertainty_dropout_p=uncertainty_dropout_p,
                    dropout_sampling_size=dropout_sampling_size,
                    n_jobs=n_jobs,
                    seed=seed,
                    continuous_fit=continuous_fit,
                    logger=logger or EmptyLogger())
    elif data_format == "graphgps":
        from graphgps.optuna.graphgps import GraphGPS
        return GraphGPS(save_dir=save_dir,
                        cfg_path=cfg_path,
                        ensemble_size=ensemble_size,
                        number_of_molecules=number_of_molecules,
                        n_jobs=n_jobs,
                        seed=seed)
    else:
        raise ValueError(f"unknown data_format {data_format}")


def get_kernel(graph_kernel_type: Literal["graph", "pre-computed", "no"] = "no",
               mgk_files: List[str] = None,
               features_kernel_type: Literal["dot_product", "rbf"] = None,
               features_hyperparameters: Union[float, List[float]] = None,
               features_hyperparameters_file: str = None,
               dataset: Dataset = None,
               kernel_pkl_path: str = None,
               ):
    if mgk_files is None:
        assert graph_kernel_type == "no"
        # no graph kernel involved.
        if features_kernel_type is None:
            return None
        elif features_kernel_type == "linear":
            raise NotImplementedError
        elif features_kernel_type == "dot_product":
            if features_hyperparameters.__class__ == list:
                assert len(features_hyperparameters) == 1
                sigma_0 = features_hyperparameters[0]
            else:
                sigma_0 = features_hyperparameters
            return DotProduct(sigma_0=sigma_0)
        elif features_kernel_type == "rbf":
            return RBF(length_scale=features_hyperparameters)
        else:
            raise ValueError
    else:
        if graph_kernel_type == "graph":
            return get_kernel_config(
                dataset=dataset,
                graph_kernel_type="graph",
                mgk_hyperparameters_files=mgk_files,
                features_kernel_type=features_kernel_type,
                features_hyperparameters=features_hyperparameters,
                features_hyperparameters_bounds="fixed",
                features_hyperparameters_file=features_hyperparameters_file
            ).kernel
        elif graph_kernel_type == "pre-computed":
            assert kernel_pkl_path is not None
            if os.path.exists(kernel_pkl_path):
                return get_kernel_config(
                    dataset=dataset,
                    graph_kernel_type="pre-computed",
                    features_kernel_type=features_kernel_type,
                    features_hyperparameters=features_hyperparameters,
                    features_hyperparameters_bounds="fixed",
                    features_hyperparameters_file=features_hyperparameters_file,
                    kernel_pkl=kernel_pkl_path
                ).kernel
            else:
                dataset.graph_kernel_type = "graph"
                kernel_config = get_kernel_config(
                    dataset=dataset,
                    graph_kernel_type="graph",
                    mgk_hyperparameters_files=mgk_files,
                    features_kernel_type=features_kernel_type,
                    features_hyperparameters=features_hyperparameters,
                    features_hyperparameters_bounds="fixed",
                    features_hyperparameters_file=features_hyperparameters_file
                )
                kernel_config = calc_precomputed_kernel_config(kernel_config=kernel_config, dataset=dataset)
                dataset.graph_kernel_type = "pre-computed"
                pickle.dump(kernel_config, open(kernel_pkl_path, "wb"), protocol=4)
                return kernel_config.kernel
        else:
            raise ValueError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import List, Literal
import os
import copy
import pandas as pd


def get_data(data_format: Literal["mgktools", "chemprop", "graphgps"],
             save_dir: str,
             file_name: str,
             smiles_columns: List[str] = None,
             targets_columns: List[str] = None,
             features_columns: List[str] = None,
             features_generators: List[str] = None,
             features_combination: Literal["concat", "mean"] = None,
             graph_kernel_type: Literal["graph", "pre-computed", "no"] = None,
             n_jobs: int = 8):
    if features_generators is not None and features_combination is None:
        features_combination = "concat"
    path = os.path.join(save_dir, file_name)
    df = pd.read_csv(path)
    if len(df) == 0:
        return None
    # Check if there are missing values in the target columns
    for targets_column in targets_columns:
        assert not df[targets_column].isnull().any(), f"Missing values in the target column {targets_column}."

    if data_format == "chemprop":
        from chemprop.data.utils import get_data
        assert features_columns is None
        dataset = get_data(path=path,
                           smiles_columns=smiles_columns,
                           target_columns=targets_columns,
                           features_generator=features_generators,
                           n_jobs=n_jobs)
    elif data_format == "graphgps":
        from graphgps.data.data_molalkit import get_data
        dataset = get_data(path=path, save_dir=save_dir,
                           smiles_columns=smiles_columns,
                           targets_columns=targets_columns,
                           features_generators=features_generators,
                           n_jobs=n_jobs)
    else:
        from mgktools.data.data import Dataset
        dataset = Dataset.from_df(df=df,
                                  smiles_columns=smiles_columns,
                                  features_columns=features_columns,
                                  targets_columns=targets_columns,
                                  n_jobs=n_jobs)
        dataset.set_status(graph_kernel_type=graph_kernel_type,
                           features_generators=features_generators, 
                           features_combination=features_combination)
        if graph_kernel_type in ["graph", "pre-computed"]:
            dataset.create_graphs(n_jobs=n_jobs)
            dataset.unify_datatype()
        if features_generators is not None:
            dataset.create_features_mol(n_jobs=n_jobs)
    if "uidx" not in df:
        df["uidx"] = range(len(df))
    for i, data in enumerate(dataset):
        data.uidx = df.iloc[i]["uidx"]
    return dataset


def get_subset_from_idx(dataset, idx):
    """ Create a subset of the dataset by index.

    Parameters
    ----------
    dataset : Chemprop or Mgktools dataset object
        The dataset to be subsetted.
    idx: List[int]
        index of the subset
    """
    subset = copy.copy(dataset)
    subset.data = [data for i, data in enumerate(dataset.data) if i in idx]
    assert len(subset) == len(idx), "Subset length does not match the index length, indicating that some indices are not found in the parent dataset."
    return subset


def get_subset_from_uidx(dataset, id2datapoint, uidx):
    """ Create a subset of the dataset by index of a dict (id2datapoint).

    Parameters
    ----------
    dataset : Chemprop or Mgktools dataset object
        The dataset to be subsetted.
    id2datapoint: Dict
        Dictionary of id to datapoint.
    idx: List[int]
        index of the subset.
    """
    dataset_ = copy.copy(dataset)
    dataset_.data = []
    for i in uidx:
        dataset_.data.append(id2datapoint[i])
    return dataset_



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import List, Union, Literal
import numpy as np
from mgktools.evaluators.metric import AVAILABLE_METRICS_BINARY, AVAILABLE_METRICS_REGRESSION, metric_binary, metric_regression


def eval_metric_func(y, y_pred, metric: str) -> float:
    if metric in AVAILABLE_METRICS_REGRESSION:
        return metric_regression(y, y_pred, metric)
    elif metric in AVAILABLE_METRICS_BINARY:
        return metric_binary(y, y_pred, metric)
    else:
        raise RuntimeError(f"Unsupported metrics {metric}")


def random_choice(N, size):
    if size < N:
        idx = np.random.choice(N, size, replace=False).tolist()
        mask = np.isin(range(N), idx)
        idx_remain = np.arange(N)[~mask].tolist()
    else:
        idx = list(range(N))
        idx_remain = []
    return idx, idx_remain


def get_topn_idx(values: np.ndarray, n: int = 1, target: Union[Literal["max", "min"], float] = "max",
                 cutoff: float = None) -> List[int]:
    """ Get the indices of top n values.

    Parameters
    ----------
    values: array-like.
    n: number of indices to be selected.
    target: "max", "min", or a float value.
    cutoff: if not None, only values >= cutoff (when target=max) will be considered.

    Returns
    -------

    """
    if isinstance(values, list):
        values = np.array(values)
    if target == "min":
        values = - values
        if cutoff is not None:
            cutoff = - cutoff
    elif isinstance(target, float):
        assert cutoff is None
        values = - np.absolute(values - target)  # distance from target
    if cutoff is not None:
        n_candidates = len(np.where(values >= cutoff)[0])
        n = min(n, n_candidates)
        if n == 0:
            return []
    # Includes tiny random values to randomly sort duplicated values
    sorting_key = values + np.random.random(len(values)) * 1e-10
    return np.argsort(sorting_key)[-n:].tolist()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from graphdot.model.gaussian_process import GaussianProcessRegressor as GPR
from molalkit.models.base import BaseModel


class GPRegressor(GPR, BaseModel):
    def __init__(self, *args, **kwargs):
        uncertainty_type = kwargs.pop("uncertainty_type")
        super().__init__(*args, **kwargs)
        self.uncertainty_type = uncertainty_type

    def fit_molalkit(self, train_data, **kwargs):
        X = train_data.X
        y = train_data.y
        super().fit(X, y, **kwargs)

    def predict_uncertainty(self, pred_data):
        X = pred_data.X
        if self.uncertainty_type == "value":
            preds = self.predict(X).reshape(-1, 1)
            preds = np.concatenate([preds, 1 - preds], axis=1)
            return (0.25 - np.var(preds, axis=1)) * 4
        elif self.uncertainty_type == "uncertainty":
            return self.predict(X, return_std=True)[1]

    def predict_value(self, pred_data):
        X = pred_data.X
        return self.predict(X)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os


class EmptyLogger:
    def debug(self, info):
        return

    def info(self, info):
        return

    def warning(self, info):
        return
        
    def error(self, info):
        return
    

def create_logger(name: str, save_dir: str = None, verbose: int = 1) -> logging.Logger:
    if verbose == 0:
        return EmptyLogger()
    
    if name in logging.root.manager.loggerDict:
        return logging.getLogger(name)

    logger = logging.getLogger(name)

    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    # Set logger depending on desired verbosity
    ch = logging.StreamHandler()
    if verbose == 2:
        ch.setLevel(logging.DEBUG)
    elif verbose == 1:
        ch.setLevel(logging.INFO)
    logger.addHandler(ch)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)

        fh_v = logging.FileHandler(os.path.join(save_dir, "%s_debug.log" % name))
        fh_v.setLevel(logging.DEBUG)
        # fh_q = logging.FileHandler(os.path.join(save_dir, "%s_quiet.log" % name))
        # fh_q.setLevel(logging.INFO)

        logger.addHandler(fh_v)
        # logger.addHandler(fh_q)

    return logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import gc
import torch
torch.cuda.is_available()
import os
import shutil
import pandas as pd
from molalkit.active_learning.learner import ActiveLearner
from molalkit.exe.args import LearningArgs


def _free_init_memory(args):
    """Free memory from data loading caches and redundant dataset references."""
    from chemprop.data.data import (
        empty_cache, set_cache_graph, set_cache_mol, set_cache_features
    )
    empty_cache()
    set_cache_graph(False)
    set_cache_mol(False)
    set_cache_features(False)
    for attr in ("_datasets_full", "_datasets_empty"):
        if hasattr(args, attr):
            delattr(args, attr)
    gc.collect()


def molalkit_run(arguments=None):
    args = LearningArgs().parse_args(arguments)
    logger = args.logger
    if args.load_checkpoint and os.path.exists("%s/al.pkl" % args.save_dir):
        logger.info("Restart active learning from checkpoint file %s/al.pkl" % args.save_dir)
        active_learner = ActiveLearner.load(path=args.save_dir)
        current_iter = active_learner.current_iter
    else:
        logger.info("Start active learning from scratch")
        active_learner = ActiveLearner(
            save_dir=args.save_dir,
            selector=args.selector,
            forgetter=args.forgetter,
            models=args.models,
            id2datapoints=args.id2datapoints,
            datasets_train=args.datasets_train,
            datasets_pool=args.datasets_pool,
            datasets_val=args.datasets_val,
            metrics=args.metrics,
            top_uidx=args.top_uidx,
            kernel=args.kernels[0],
            detail=args.detail,
        )
        current_iter = 0
        active_learner.evaluate()
    _free_init_memory(args)
    for i in range(current_iter, args.max_iter or 100):
        logger.info("Active learning loop %d" % i)
        for _ in range(args.n_select):
            active_learner.step_select()
            logger.debug("Select step %d" % _)
        if args.f_min_train_size is None or len(active_learner.datasets_train[0]) >= args.f_min_train_size:
            for _ in range(args.n_forget):
                active_learner.step_forget()
                logger.debug("Forget step %d" % _)
        if args.evaluate_stride is not None and i % args.evaluate_stride == 0:
            active_learner.evaluate()
            logger.debug("Evaluate step")
        if i % args.write_traj_stride == 0:
            active_learner.write_traj()
        if args.save_cpt_stride is not None and i % args.save_cpt_stride == 0:
            active_learner.current_iter = i + 1
            active_learner.save(path=args.save_dir, filename="al_temp.pkl", overwrite=True)
            shutil.move(os.path.join(args.save_dir, "al_temp.pkl"), os.path.join(args.save_dir, "al.pkl"))
            logger.info("Save checkpoint file %s/al.pkl" % args.save_dir)
    df = pd.read_csv(f"{args.save_dir}/full.csv")
    df[df["uidx"].isin([data.uidx for data in active_learner.datasets_train[0]])].to_csv(f"{args.save_dir}/train_end.csv", index=False)
    df[df["uidx"].isin([data.uidx for data in active_learner.datasets_pool[0]])].to_csv(f"{args.save_dir}/pool_end.csv", index=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Literal
from tqdm import trange
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
from torch.nn.utils.rnn import pad_sequence
from mgktools.data.data import Dataset
from molalkit.models.rnn.data import MoleculeDataset
from molalkit.models.rnn.tokenizer import BaseTokenizer


class MoleculeRNN(nn.Module):
    def __init__(self,
                 task_type: Literal["binary", "regression"],
                 vocab_size: int,
                 rnn_type: Literal["LSTM", "GRU"] = "GRU",
                 embedding_dim: int = 64,
                 depth: int = 2,
                 hidden_size: int = 128,
                 dropout: float = 0.0,
                 ffn_num_layers: int = 3):
        super(MoleculeRNN, self).__init__()
        self.task_type = task_type
        self.embedding = nn.Embedding(vocab_size + 1, embedding_dim)
        self.rnn_type = rnn_type
        self.depth = depth
        self.hidden_size = hidden_size

        if self.rnn_type == "LSTM":
            self.rnn = nn.LSTM(
                embedding_dim,
                hidden_size,
                depth,
                batch_first=True,
            )
        elif self.rnn_type == "GRU":
            self.rnn = nn.GRU(
                embedding_dim,
                hidden_size,
                depth,
                batch_first=True,
            )
        else:
            raise ValueError("rnn_type must be either LSTM or GRU")
        
        dropout = nn.Dropout(dropout)
        activation = nn.ReLU()
        ffn = []
        for i in range(ffn_num_layers):
            if i == ffn_num_layers - 1:
                ffn += [
                    dropout,
                    nn.Linear(hidden_size, 1)
                ]
            else:
                ffn += [
                    dropout,
                    nn.Linear(hidden_size, hidden_size),
                    activation
                ]
        self.ffn = nn.Sequential(*ffn)

        if self.task_type == "binary":
            self.sigmoid = nn.Sigmoid()
        # Initialize weights
        for param in self.parameters():
            if param.dim() > 1:
                nn.init.xavier_uniform_(param)
            else:
                nn.init.zeros_(param)

    def forward(self, x):
        batch_size = x.shape[0]
        # x shape: (batch_size, sequence_length)
        embedded = self.embedding(x)
        # embedded shape: (batch_size, sequence_length, embedding_dim)
        if self.rnn_type == "LSTM":
            h0 = torch.zeros(self.depth, batch_size, self.hidden_size).to(x.device)
            c0 = torch.zeros(self.depth, batch_size, self.hidden_size).to(x.device)
            output, _ = self.rnn(embedded, (h0, c0))
        elif self.rnn_type == "GRU":  # GRU
            h0 = torch.zeros(self.depth, batch_size, self.hidden_size).to(x.device)
            output, _ = self.rnn(embedded, h0)
        
        output = output[: ,-1]
        # return self.ffn(output)
        if self.task_type == "binary":
            return self.sigmoid(self.ffn(output))
        else:
            return self.ffn(output)


class RNN:
    def __init__(self, tokenizer: BaseTokenizer,
                 task_type: Literal["binary", "regression"], rnn_type: Literal["LSTM", "GRU"] = "LSTM",
                 embedding_size: int = 128, hidden_size: int = 256, depth: int = 2, dropout: float = 0.1,
                 ffn_num_layers: int = 3, epochs: int = 30, batch_size: int = 64):
        self.tokenizer = tokenizer
        self.task_type = task_type
        self.rnn_type = rnn_type
        self.embedding_dim = embedding_size
        self.depth = depth
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.ffn_num_layers = ffn_num_layers
        self.epochs = epochs
        self.batch_size = batch_size
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    
    def fit_molalkit(self, train_data: Dataset, iteration: int = 0):
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        # initialize model
        model = MoleculeRNN(
            task_type=self.task_type,
            vocab_size=len(self.tokenizer.vocab),
            rnn_type=self.rnn_type,
            embedding_dim=self.embedding_dim,
            depth=self.depth,
            hidden_size=self.hidden_size,
            dropout=self.dropout,
            ffn_num_layers=self.ffn_num_layers
        )
        model = model.to(self.device)

        # initialize data loader
        train_data_loader = self.get_dataloader(train_data)

        if self.task_type == "binary":
            loss = nn.BCELoss()
        else:
            loss = nn.MSELoss()
        optimizer = torch.optim.AdamW(model.parameters(), lr=1e-3)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer,
                                                                         T_0=5,
                                                                         T_mult=2)
        for epoch in trange(self.epochs):
            model.train()
            epoch_loss = 0.
            for batch in train_data_loader:
                x, y = batch
                x = x.to(self.device)
                y = y.to(self.device)
                optimizer.zero_grad()
                y_pred = model(x)
                batch_loss = loss(y_pred, y.unsqueeze(1))
                batch_loss.backward()
                optimizer.step()
                scheduler.step()
                epoch_loss += batch_loss.item()
            # print(epoch_loss / len(train_data_loader))
        self.model = model

    def predict_value(self, pred_data):
        pred_data_loader = self.get_dataloader(pred_data)
        with torch.no_grad():
            self.model.eval()
            y_pred = []
            for batch in pred_data_loader:
                x, _ = batch
                x = x.to(self.device)
                y_pred.append(self.model(x).cpu().numpy())
            y_pred = np.concatenate(y_pred)
            return y_pred.ravel()

    def predict_uncertainty(self, pred_data):
        assert self.task_type == "binary", "Uncertainty estimation is only available for binary classification"
        preds = self.predict_value(pred_data)
        preds = np.array([preds, 1-preds]).T
        return (0.25 - np.var(preds, axis=1)) * 4

    def get_dataloader(self, data: Dataset):
        assert data.X_smiles.shape[1] == 1, "Only single-column SMILES data is supported for RNN models."
        assert data.y.shape[1] == 1, "Only single-column target data is supported for RNN models."
        data_ = MoleculeDataset(
            smiles_list=data.X_smiles.ravel().tolist(),
            targets=data.y.ravel().tolist(),
            tokenizer=self.tokenizer
        )

        def pad_collate(batch):
            """
            Put the sequences of different lengths in a minibatch by paddding.
            """
            # embedding layer takes long tensors
            x = [torch.tensor(x[0], dtype=torch.long) for x in batch]
            y = [torch.tensor(x[1], dtype=torch.float) for x in batch]

            x_padded = pad_sequence(
                x, 
                batch_first=True,
                padding_value=self.tokenizer.vocab["<pad>"]
            )

            return x_padded, torch.stack(y)
        
        return DataLoader(
            data_, 
            batch_size=self.batch_size, 
            shuffle=False,
            collate_fn=pad_collate
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from typing import List, Dict, Optional, Callable
import numpy as np
import pandas as pd
from mgktools.features_mol.features_generators import FeaturesGenerator
from molalkit.models.utils import get_kernel, get_model
from molalkit.models.configs import MODEL_DIR


def get_features_generators_from_config(model_config: Dict) -> Optional[List[List[FeaturesGenerator]]]:
    fingerprints_class = model_config.get("fingerprints_class")
    radius = model_config.get("radius")
    num_bits = model_config.get("num_bits")
    atomInvariantsGenerator = model_config.get("atomInvariantsGenerator")
    if fingerprints_class is None:
        return None
    else:
        return [FeaturesGenerator(features_generator_name=fc,
                                  radius=radius,
                                  num_bits=num_bits,
                                  atomInvariantsGenerator=atomInvariantsGenerator) for fc in fingerprints_class]


def get_kernel_from_config(model_config: Dict, dataset, kernel_pkl_path) -> Callable:
    return get_kernel(
        graph_kernel_type=model_config.get("graph_kernel_type") or "no",
        mgk_files=model_config.get("mgk_files"),
        features_kernel_type=model_config.get("features_kernel_type"),
        features_hyperparameters=model_config.get("features_hyperparameters"),
        features_hyperparameters_file=model_config.get("features_hyperparameters_file"),
        dataset=dataset,
        kernel_pkl_path=kernel_pkl_path,
    )


def _resolve_cfg_path(cfg):
    """Resolve a YAML config path, checking MODEL_DIR if not found directly."""
    if cfg is None:
        return None
    if os.path.exists(cfg):
        return cfg
    resolved = os.path.join(MODEL_DIR, cfg)
    if os.path.exists(resolved):
        return resolved
    return cfg


def get_model_from_config(model_config: Dict, dataset, task_type, save_dir,
                          data_path, smiles_columns, targets_columns, 
                          features_generators, kernel,
                          n_jobs, seed, logger) -> Callable:
    data_format = model_config["data_format"]
    if data_format == "mgktools":
        smiles_full = dataset.X_smiles.ravel()
        features_size = None
    elif data_format == "graphgps":
        smiles_full = None
        features_size = None
    else:
        smiles_full = None
        features_size = dataset.features_size()
    return get_model(
        data_format=model_config["data_format"],
        task_type=task_type,
        model=model_config.get("model"),
        kernel=kernel,
        uncertainty_type=model_config.get("uncertainty_type"),
        alpha=model_config.get("alpha"),
        C=model_config.get("C"),
        booster=model_config.get("booster"),
        n_estimators=model_config.get("n_estimators") or 100,
        max_depth=model_config.get("max_depth"),
        learning_rate=model_config.get("learning_rate") or 0.1,
        tokenizer=model_config.get("tokenizer"),
        smiles_full=smiles_full,
        embedding_size=model_config.get("embedding_size") or 64,
        save_dir=save_dir,
        data_path=data_path,
        smiles_columns=smiles_columns,
        target_columns=targets_columns,
        loss_function=model_config.get("loss_function"),
        multiclass_num_classes=model_config.get("loss_function") or 3,
        features_generator=features_generators,
        no_features_scaling=model_config.get("no_features_scaling") or False,
        features_only=model_config.get("features_only") or False,
        features_size=features_size,
        epochs=model_config.get("epochs") or 50,
        depth=model_config.get("depth") or 3,
        hidden_size=model_config.get("hidden_size") or 300,
        ffn_num_layers=model_config.get("ffn_num_layers") or 2,
        ffn_hidden_size=model_config.get("ffn_hidden_size"),
        dropout=model_config.get("dropout") or 0.0,
        batch_size=model_config.get("batch_size") or 50,
        ensemble_size=model_config.get("ensemble_size") or 1,
        number_of_molecules=model_config.get("number_of_molecules") or len(smiles_columns),
        mpn_shared=model_config.get("mpn_shared") or False,
        atom_messages=model_config.get("atom_messages") or False,
        undirected=model_config.get("undirected") or False,
        const_lr=model_config.get("const_lr") or False,
        weight_decay=model_config.get("weight_decay") or 0.0,
        cbp=model_config.get("cbp") or False,
        replacement_rate=model_config.get("replacement_rate") or 0.00001,
        maturity_threshold=model_config.get("maturity_threshold") or 100,
        reinit_weights=model_config.get("reinit_weights") or 'xavier',
        decay_rate=model_config.get("decay_rate") or 0.99,
        class_balance=model_config.get("class_balance") or False,
        checkpoint_dir=model_config.get("checkpoint_dir"),
        checkpoint_frzn=model_config.get("checkpoint_frzn"),
        frzn_ffn_layers=model_config.get("frzn_ffn_layers") or 0,
        freeze_first_only=model_config.get("freeze_first_only") or False,
        mpn_path=model_config.get("mpn_path"),
        freeze_mpn=model_config.get("freeze_mpn") or False,
        uncertainty_method=model_config.get("uncertainty_method"),
        uncertainty_dropout_p=model_config.get("uncertainty_dropout_p") or 0.1,
        dropout_sampling_size=model_config.get("dropout_sampling_size") or 10,
        continuous_fit=model_config.get("continuous_fit") or False,
        cfg_path=_resolve_cfg_path(model_config.get("cfg")),
        pretrained_path=model_config.get("pretrained_path"),
        n_head=model_config.get("n_head") or 12,
        n_layer=model_config.get("n_layer") or 12,
        n_embd=model_config.get("n_embd") or 768,
        d_dropout=model_config.get("d_dropout") or 0.1,
        num_feats=model_config.get("num_feats") or 32,
        n_jobs=n_jobs,
        seed=seed,
        logger=logger
    )


def read_df(path, task_type, target_columns) -> pd.DataFrame:
    df = pd.read_csv(path)
    if task_type in ["binary", "multiclass"]:
        for target_column in target_columns:
            if df[target_column].dtype == "float":
                df[target_column] = df[target_column].astype(int)
    return df


def add_error_rate_to_labels(df, error_rate, target_column):
    """ Randomly flip the labels of a portion of the data."""
    error_index = np.random.choice(df.index, int(error_rate * len(df)), replace=False)
    df["flip_label"] = False
    df.loc[error_index, target_column] ^= 1
    df.loc[error_index, "flip_label"] = True


def apply_error_rate_to_id2datapoint(id2datapoint, df, target_column):
    assert "flip_label" in df.columns, "Error rate has not been applied to the dataframe."
    for i, row in df.iterrows():
        if row["flip_label"]:
            assert set([id2datapoint[row["uidx"]].targets[0], row[target_column]]) == {0, 1}
            id2datapoint[row["uidx"]].targets[0] = row[target_column]
        else:
            assert id2datapoint[row["uidx"]].targets[0] == row[target_column]
